import os

import sys
import itertools
import numpy as np
import subprocess
from shutil import which

docstring = """
Use this script like this

Or get the data into a file which list the data of the edges like:
weight vertex1 mode1 vertex2 mode2
weight vertex1 mode1 vertex2 mode2

python leiwand.py in=filename

You can call the script with several options like this
python leiwand.py in=filename key1=value1 key2=value2

option keys are the following
"""

variables = {
    "line_width": 4.0,
    "radius": 3,
    "col0": "{RGB}{0,0,204}",
    "col1": "{RGB}{204,0,0}",
    "col2": "{RGB}{0,204,0}",
    "col3": "{RGB}{255, 165, 0}",
    "col4": "{RGB}{128,0,128}",
    "col5": "{RGB}{255, 255, 0}",
    "col6": "{RGB}{102, 0, 102}",
    "vertexcolor": "{RGB}{250,250,250}",
    "fontcolor": "{RGB}{0,0,0}",
    "angle": 0,
    "vertices": None,
    "whitespace": "10pt"
}


def leiwand(data, name='graph'):
    poly = {}
    output = name
    numcolors = 7

    external_vertices = None
    if variables['vertices'] is not None:
        external_vertices = variables["vertices"].split(' ')
        # reverse order (drawing is counter-clockwise)
        external_vertices = list(reversed(external_vertices))
        # print("got vertices: ", external_vertices)

    whitespace = None
    if variables["whitespace"] is not None:
        whitespace = variables["whitespace"]
    with open(output + ".tex", "w") as outf:
        optionmap = {tuple([c1, c2, True]): f"bicolor={{col{c1}}}{{col{c2}}}" for c1, c2 in
                     itertools.permutations(range(numcolors), 2)}
        for ii in range(numcolors):
            optionmap[(ii, ii, True)] = f"bicolor={{col{ii}}}{{col{ii}}}"
        optionmap.update({tuple([c1, c2, False]): f"bicolor_neg={{col{c2}}}{{col{c1}}}" for c1, c2 in
                          itertools.product(range(numcolors), repeat=2)})
        # if use f"bicolor_neg={{col{c1}}}{{col{c2}}}" will inverse the color for the negitive edges, don't know why; 
        # print(optionmap)
        if whitespace is not None:
            print("\documentclass[border={}]{}".format(whitespace, r"{standalone}"), file=outf)
        else:
            print(r"\documentclass{standalone}", file=outf)
        print(r"""
        
         \usepackage{tikz}
         \usetikzlibrary{decorations.markings, shapes.geometric} 
         
        """, file=outf)

        colors = r"\definecolor{vertexcol}" + variables["vertexcolor"]
        for ii in range(numcolors):
            colors += f"\definecolor{{col{ii}}}" + variables[f"col{ii}"]
        colors += r"\definecolor{fontcolor}" + variables["fontcolor"]
        print(colors, file=outf)

        print(r"""
        \begin{document}
      
        \tikzset{
            bicolor/.style n args={2}{
                postaction={draw=#1, decoration={curveto,  post=moveto, post length=0.5*\pgfmetadecoratedpathlength}, decorate},
                postaction={draw=#2, decoration={curveto,  pre=moveto, pre length=0.5*\pgfmetadecoratedpathlength}, decorate},
            },
            bicolor_neg/.style n args={2}{
                postaction={draw=#1, decoration={curveto,  pre=moveto, pre length=0.5*\pgfmetadecoratedpathlength}, decorate},
                postaction={draw=#2, decoration={curveto,  post=moveto, post length=0.5*\pgfmetadecoratedpathlength}, decorate},
                postaction={decoration={markings, mark=at position 0.5 with {\node[diamond, fill=white, aspect=0.7,draw, thin, minimum width=10pt,opacity=1] {};}}, decorate},
            },
            vertex/.style={circle, draw=black, ultra thick ,fill=vertexcol!80,minimum size=15pt},
        }

        \begin{tikzpicture}
        
        """, file=outf)

        vertices = []
        weights = []
        for d in data:
            weights.append(abs(d[0]))
            vertices.append(d[1])
            vertices.append(d[3])
        vertices = list(set(vertices))
        max_weight = max(weights)

        # check if vertices where specified manually
        if external_vertices is not None:
            print("replacing: ", vertices)
            print("with external vertices: ", external_vertices)
            vertices = external_vertices
        else:
            # sort vertices alphabetically
            vertices = list(sorted(vertices))

        if len(poly) < len(vertices):
            # poly = Polygon.regular(len(vertices), radius=variables["radius"], angle=float(180))
            angles = [2 * np.pi * ii / len(vertices) for ii in range(len(vertices))]
            poly = [tuple([variables["radius"] * np.cos(theta - variables["angle"]),
                           variables["radius"] * np.sin(theta - variables["angle"])]) for theta in angles]
        else:
            # sort alphabetically
            poly = reversed(list(dict(sorted(poly.items(), key=lambda x: x[0])).values()))
        for i, coord in enumerate(poly):
            print(r"\node[vertex] ({name}) at ({x},{y}) {xname};".format(name=vertices[i],
                                                                         xname=r"{\color{fontcolor}" + vertices[
                                                                             i] + "}", x=coord[0], y=coord[1]),
                  file=outf)

        # edge_string = r"\path ({v1}) edge[{options}, opacity={opacity}] ({v2});"
        edge_string = r"\path[{options}, opacity={opacity}] ({v1}) to ({v2});"

        for d in data:
            assert (len(d) == 6)
            weight = d[0]
            v1 = d[1]
            t1 = d[2]
            v2 = d[3]
            t2 = d[4]
            b = d[5]
            opacity = max(0.3, abs(weight) / max_weight)
            angles = [360 * ii / len(vertices) for ii in range(len(vertices))]
            weight_pos = (weight > 0)
            if v1 == v2:  # loop
                angle1 = angles[int(v1)] + 30
                angle2 = angles[int(v1)] - 30

                print(edge_string.format(v1=v1, v2=v2,
                                         options="line width={lw},".format(lw=variables["line_width"]) + optionmap[
                                             (t1, t2, weight_pos)] + ", looseness=" + str(
                                             b) + f",right,out={angle1},in={angle2}",
                                         opacity=opacity), file=outf)
            else:
                print(edge_string.format(v1=v1, v2=v2,
                                         options="line width={lw},".format(lw=variables["line_width"]) + optionmap[
                                             (t1, t2, weight_pos)] + ", bend right=" + str(b),
                                         opacity=opacity), file=outf)
        print(r"""
        \end{tikzpicture}

        \end{document}
        """, file=outf)

    # Get the current working directory
    cwd = os.getcwd()

    # Print the current working directory
    print("Current working directory: {0}".format(cwd))

    print("created {}.tex".format(output))
    print("trying to compile with pdflatex ... might be caught in endless loop")

    import subprocess
    from shutil import which

    system_has_pdflatex = which("pdflatex") is not None
    if not system_has_pdflatex:
        raise Exception("You need pdflatex in order to export circuits to pdfs")

    with open(output + ".log", "w") as file:
        subprocess.call(["pdflatex", output + ".tex"], stdout=file)

    print("created {}.pdf".format(output))


def leiwandBulk(data, cnfg, name='graph', root=""):
    #go into directory where graph should be saved
    os.chdir(os.path.join(os.getcwd(), root))
    poly = {}
    output = name
    numcolors = 7
    # defining shapes for different kinds of vertices
    shape_dict = {'in': 'inner sep=0.1em, regular polygon,regular polygon sides=3', 'out': "circle", "anc": "rectangle", 'sps': 'inner sep=0.1em,regular polygon,regular polygon sides=3',
                  'mix': 'star,star points=10'}

    external_vertices = None
    if variables['vertices'] is not None:
        external_vertices = variables["vertices"].split(' ')
        # reverse order (drawing is counter-clockwise)
        external_vertices = list(reversed(external_vertices))
        # print("got vertices: ", external_vertices)

    whitespace = None
    if variables["whitespace"] is not None:
        whitespace = variables["whitespace"]
    with open(output + ".tex", "w") as outf:
        optionmap = {tuple([c1, c2, True]): f"bicolor={{col{c1}}}{{col{c2}}}" for c1, c2 in
                     itertools.permutations(range(numcolors), 2)}
        for ii in range(numcolors):
            optionmap[(ii, ii, True)] = f"bicolor={{col{ii}}}{{col{ii}}}"
        optionmap.update({tuple([c1, c2, False]): f"bicolor_neg={{col{c2}}}{{col{c1}}}" for c1, c2 in
                          itertools.product(range(numcolors), repeat=2)})
        # if use f"bicolor_neg={{col{c1}}}{{col{c2}}}" will inverse the color for the negitive edges, don't know why;
        # print(optionmap)
        if whitespace is not None:
            print("\documentclass[border={}]{}".format(whitespace, r"{standalone}"), file=outf)
        else:
            print(r"\documentclass{standalone}", file=outf)
        print(r"""

         \usepackage{tikz}
         \usetikzlibrary{decorations.markings, shapes.geometric} 

        """, file=outf)

        colors = r"\definecolor{vertexcol}" + variables["vertexcolor"]
        for ii in range(numcolors):
            colors += f"\definecolor{{col{ii}}}" + variables[f"col{ii}"]
        colors += r"\definecolor{fontcolor}" + variables["fontcolor"]
        print(colors, file=outf)

        print(r"""
        \begin{document}

        \tikzset{
            bicolor/.style n args={2}{
                postaction={draw=#1, decoration={curveto,  post=moveto, post length=0.5*\pgfmetadecoratedpathlength}, decorate},
                postaction={draw=#2, decoration={curveto,  pre=moveto, pre length=0.5*\pgfmetadecoratedpathlength}, decorate},
            },
            bicolor_neg/.style n args={2}{
                postaction={draw=#1, decoration={curveto,  pre=moveto, pre length=0.5*\pgfmetadecoratedpathlength}, decorate},
                postaction={draw=#2, decoration={curveto,  post=moveto, post length=0.5*\pgfmetadecoratedpathlength}, decorate},
                postaction={decoration={markings, mark=at position 0.5 with {\node[diamond, fill=white, aspect=0.7,draw, thin, minimum width=10pt,opacity=1] {};}}, decorate},
            },
            vertex/.style={circle, draw=black, ultra thick ,fill=vertexcol!80,minimum size=15pt},
        }

        \begin{tikzpicture}

        """, file=outf)

        vertices = []
        weights = []
        for d in data:
            weights.append(abs(d[0]))
            vertices.append(d[1])
            vertices.append(d[3])
        vertices = list(set(vertices))
        max_weight = max(weights)

        # check if vertices where specified manually
        if external_vertices is not None:
            print("replacing: ", vertices)
            print("with external vertices: ", external_vertices)
            vertices = external_vertices
        else:
            # sort vertices alphabetically
            vertices = list(sorted(vertices))

        if len(poly) < len(vertices):
            # poly = Polygon.regular(len(vertices), radius=variables["radius"], angle=float(180))
            angles = [2 * np.pi * ii / len(vertices) for ii in range(len(vertices))]
            poly = [tuple([variables["radius"] * np.cos(theta - variables["angle"]),
                           variables["radius"] * np.sin(theta - variables["angle"])]) for theta in angles]
        else:
            # sort alphabetically
            poly = reversed(list(dict(sorted(poly.items(), key=lambda x: x[0])).values()))
        for i, coord in enumerate(poly):
            print(r"\node[vertex] ({name}) at ({x},{y}) [{shape}] {xname};".format(name=vertices[i], shape=shape_dict[
                cnfg["vert_types"][i]], xname=r"{\color{fontcolor}" + vertices[i] + "}", x=coord[0], y=coord[1]),
                  file=outf)

        # edge_string = r"\path ({v1}) edge[{options}, opacity={opacity}] ({v2});"
        edge_string = r"\path[{options}, opacity={opacity}] ({v1}) to ({v2});"

        for d in data:
            assert (len(d) == 6)
            weight = d[0]
            v1 = d[1]
            t1 = d[2]
            v2 = d[3]
            t2 = d[4]
            b = d[5]
            opacity = max(0.3, abs(weight) / max_weight)
            angles = [360 * ii / len(vertices) for ii in range(len(vertices))]
            weight_pos = (weight > 0)
            if v1 == v2:  # loop
                angle1 = angles[int(v1)] + 30
                angle2 = angles[int(v1)] - 30

                print(edge_string.format(v1=v1, v2=v2,
                                         options="line width={lw},".format(lw=variables["line_width"]) + optionmap[
                                             (t1, t2, weight_pos)] + ", looseness=" + str(
                                             b) + f",right,out={angle1},in={angle2}",
                                         opacity=opacity), file=outf)
            else:
                print(edge_string.format(v1=v1, v2=v2,
                                         options="line width={lw},".format(lw=variables["line_width"]) + optionmap[
                                             (t1, t2, weight_pos)] + ", bend right=" + str(b),
                                         opacity=opacity), file=outf)
        print(r"""
        \end{tikzpicture}

        \end{document}
        """, file=outf)

    # Get the current working directory
    cwd = os.getcwd()

    # Print the current working directory
    # print("Current working directory: {0}".format(cwd))

    # print("created {}.tex".format(output))
    # print("trying to compile with pdflatex ... might be caught in endless loop")

    system_has_pdflatex = which("pdflatex") is not None
    if not system_has_pdflatex:
        raise Exception("You need pdflatex in order to export circuits to pdfs")

    with open(output + ".log", "w") as file:
        subprocess.call(["pdflatex", output + ".tex"], stdout=file)

    print("created {}.pdf".format(output))


if __name__ == "__main__":
    print(docstring)
    for k, v in variables.items():
        print("{}={}".format(k, v))

    print("calling with: ", sys.argv)
    leiwand(sys.argv)



import os
import glob


def find_nth(haystack, needle, n):
    start = haystack.find(needle)
    while start >= 0 and n > 1:
        start = haystack.find(needle, start+len(needle))
        n -= 1
    return start


origin = os.getcwd()
origin

path_category = glob.glob('HighlyEntangledStates/*/*.json') # this already ignores the folders 'other_solutions'

path_category

for path in sorted(path_category):
    slash1 = find_nth(path,'/',1)
    slash2 = find_nth(path,'/',2)
    folder_name = path[ slash1 +1: slash2]
    total_path = origin + '/' + path 
    if ('/plot_' in path) or ('/config_' in path):
        pass
    else: 
        if ('rough' in path) or ('clean' in path):
            new_path = total_path.replace(folder_name+'/',f'{folder_name}/plot_{folder_name}_')
        else:
            new_path = total_path.replace(folder_name+'/',f'{folder_name}/config_')
        os.rename(total_path,new_path)



import unittest
from pathlib import Path

from tests.fast.config import GHZ_346, BELL
from pytheus.fancy_classes import Graph, defaultValues
from pytheus.main import run_main, read_config


class TestFancyClassesModule(unittest.TestCase):

    def test_defaultValues_not_imaginary(self):
        actual = defaultValues(87, False)
        self.assertEqual([True] * 87, actual)

    def test_defaultValues_cartesian(self):
        actual = defaultValues(10, 'cartesian')
        self.assertEqual([True] * 10, actual)

    def test_defaultValues_polar(self):
        actual = defaultValues(2, 'polar')
        self.assertEqual([(True, False), (True, False)], actual)

    def test_defaultValues_raises_exception_on_invalid_input(self):
        with self.assertRaises(ValueError):
            defaultValues(10, 'blabla')


class TestGraph(unittest.TestCase):
    def test_init_graph_edges_given_not_imaginary_rest_default(self):
        graph = Graph(BELL['edges'], imaginary=False)

        expected_graph_attribute = {(0, 1, 0, 0): True, (0, 1, 1, 1): True}
        self.assertEqual(expected_graph_attribute, graph.graph)
        self.assertEqual([((0, 1, 0, 0),), ((0, 1, 1, 1),)], graph.perfect_matchings)
        self.assertEqual({
            ((0, 0), (1, 0)): [((0, 1, 0, 0),)], ((0, 1), (1, 1)): [((0, 1, 1, 1),)]
        }, graph.state_catalog)






#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: cruizgo, soerenarlt, janpe
"""
import itertools
from math import factorial
from collections import Counter
import random
import numpy as np


# # Auxiliary Functions
# Helpful functions used in many processes but not by the final user. 

def allPairSplits(lst):
    '''     
    Generate all sets of unique pairs from a list `lst`.
    This is equivalent to all partitions of `lst` (considered as an indexed set) which have 2 elements
    in each partition.
    
    Recall how we compute the total number of such partitions. Starting with a list [1, 2, 3, 4, 5, 6]
    one takes off the first element, and chooses its pair [from any of the remaining 5]. For example, 
    we might choose our first pair to be (1, 4). Then, we take off the next element, 2, and choose 
    which element it is paired to (say, 3). So, there are 5 * 3 * 1 = 15 such partitions.
    That sounds like a lot of nested loops (i.e. recursion), because 1 could pick 2, in which case our 
    next element is 3. But, if one abstracts "what the next element is", and instead just thinks of what 
    index it is in the remaining list, our choices are static and can be aided by the product function.
    
    From selfgatoatigrado: https://stackoverflow.com/a/13020502
    '''
    N = len(lst)
    choice_indices = itertools.product(*[range(k) for k in reversed(range(1, N, 2))])

    for choice in choice_indices:
        # calculate the list corresponding to the choices
        tmp = lst[:]
        result = []
        for index in choice:
            result.append((tmp.pop(0), tmp.pop(index)))
        yield result  # use yield and then turn it into a list is faster than append


def targetEdges(nodes, graph):
    '''
    Returns all graph's edges which connect to the input nodes.
    '''
    return [edge for edge in graph if (edge[0] in nodes) or (edge[1] in nodes)]


def removeNodes(nodes, graph):
    '''
    Removes all graph's edges which connect to the input nodes.
    '''
    return [edge for edge in graph if not ((edge[0] in nodes) or (edge[1] in nodes))]


def deadEndEdges(graph):  # Unused
    '''
    Returns all edges connecting nodes with degree one.
    '''
    unq, counts = np.unique(np.array(graph)[:, :2], return_counts=True)
    return targetEdges(unq[counts == 1], graph)


def edgeBleach(color_edges):
    '''
    Given a list of colored edges, returns a dictionary with the available colors (values) for each 
    of the edges (keys).
    
    This function helps to speed up the function `findPerfectMatchings`.
    
    Parameters
    ----------
    color_edges : list of tuples
        List of all colored edges: [(node1, node2, color1, color2), ...].
        
    Returns
    -------
    bleached_edges : dictionary
        Dictionary with the color combinations available for each edge:
        {(node1, node2): [(color1, color2), ...], (node1, node3): [(color2, color1), ...], ...}
    '''
    raw_edges = np.unique(np.array(color_edges)[:, :2], axis=0)
    bleached_edges = {tuple(edge): [] for edge in raw_edges}
    for edge in color_edges:
        bleached_edges[edge[:2]].append(edge[2:])
    return bleached_edges


def allColorGraphs(color_nodes, loops=False):
    '''
    Given a list of colored nodes, i.e., an state. It uses AllPairSplits to generate all graphs that
    leads to such state. This function is a building block of the function 'allEdgeCovers'.
    
    Parameters
    ----------
    color_nodes : list of tuples
        List of all colored nodes: [(node,color)...] Some may be repeated.
    loops : boolean, optional
        Allow edges to connect twice the same node: (node1, node1, color1, color2).
        
    Returns
    -------
    graph_list : list of tuples
        Nested list with all graphs that produce a given state.
        Example: given the nodes [(0,0),(0,1),(1,1),(1,1)] it produces:
            [[(0, 0, 0, 1), (1, 1, 1, 1)],    - only if loops are allowed -
             [(0, 1, 0, 1), (0, 1, 1, 1)]]
    '''
    color_graph = list(allPairSplits(sorted(list(color_nodes))))
    for graph in color_graph: graph.sort()
    if loops:
        color_graph = [[[nd[0][0], nd[1][0], nd[0][1], nd[1][1]] for nd in graph]
                       for graph in color_graph]
    else:
        color_graph = [[[nd[0][0], nd[1][0], nd[0][1], nd[1][1]] for nd in graph
                        if nd[0][0] != nd[1][0]] for graph in color_graph]
        color_graph = [graph for graph in color_graph if len(graph) == len(color_nodes) / 2]
    return [tuple(tuple(ed) for ed in graph) for graph in np.unique(color_graph, axis=0)]



# # Informative Functions
# Return some information about dimensions, edges, graphs...

def nodeDegrees(edge_list, nodes_list=[], increasing=True):
    '''
    Compute the degree of each node of a graph. Returning a list of tuples such that: 
    [(node1, degree1), (node2, degree2) ...]
    By default, it sorts the nodes by their degree in increasing order.
    
    Parameters
    ----------
    edge_list : list of tuples
        List of all available edges: [(node1, node2, color1, color2), ...]
    nodes_list : list, optional
        List of all nodes. By default, the list is obtained from the nodes that appear in edge_list.
    increasing : boolean, optional
        If True, the nodes are ordered by degree.
    
    Returns
    -------
    links : list of tuples
        List of nodes and their degrees: [(node1, degree1), ...]
    '''
    if len(nodes_list) == 0: nodes_list = np.unique(np.array(edge_list)[:, :2])
    links = {ii: 0 for ii in nodes_list}
    for edge in edge_list:
        links[edge[0]] += 1
        links[edge[1]] += 1
    if increasing:
        return sorted(links.items(), key=lambda item: item[1])
    else:
        return [(k, v) for k, v in links.items()]


def graphDimensions(edge_list):
    '''
    Estimate the dimensions of a graph based on its edges.

    The output dimensions are in decreasing order: we assume the dimension of a node N, is equal or
    larger than the dimension of a node N+1.
    
    Parameters
    ----------
    edge_list : list of tuples
        List of all available edges: [(node1, node2, color1, color2), ...]
    
    Returns
    -------
    dimensions : list
        List with the dimensions accessible to each node in decreasing order. 
    '''
    color_nodes = set()
    for edge in edge_list:
        color_nodes.add((edge[0], edge[2]))
        color_nodes.add((edge[1], edge[3]))
    color_nodes = sorted(color_nodes, reverse=True)
    max_node = color_nodes[0][0]
    max_dim = color_nodes[0][1]
    dimensions = np.array([max_dim + 1] * (max_node + 1))
    for node in color_nodes:
        if node[1] > max_dim:
            max_dim = node[1]
            dimensions[:(node[0] + 1)] = max_dim + 1
    return list(dimensions)


def stateDimensions(ket_list):
    '''
    Given a state, it returns the dimensions accessible to each particle.
    
    Parameters
    ----------
    ket_list : list of tuples
        List of kets written as: [((node1, dimension1), (node2, dimension2), ...), ...].
    
    Returns
    -------
    dimensions : list
        List with the dimensions accessible to each particle. 
    '''
    dimensions = [1] * len(ket_list[0])
    for ket in ket_list:
        for ii, local_dim in enumerate(ket):
            dimensions[ii] = np.maximum(dimensions[ii], local_dim[1] + 1)
    return dimensions


# # Generators
# Used to produce edges, graphs, states...

def stateCatalog(graph_list):
    '''
    Given a list of colored graphs, returns a dictionary with graphs grouped by the state they generate.
    Each of these states is a dictionary key.
    
    Parameters
    ----------
    graph_list : list
        List of graphs.
    
    Returns
    -------
    state_catalog : dictionary
        Dictionary with all the kets created by each graph as keys. If more than one graph leads to
        the same state, they are listed together in the corresponding entrance of the dictionary.
    '''
    state_catalog = {}
    for graph in graph_list:
        coloring = []
        for ed in graph: coloring += [(ed[0], ed[2]), (ed[1], ed[3])]
        coloring = tuple(sorted(coloring))
        try:
            state_catalog[coloring] += [graph]
        except KeyError:
            state_catalog[coloring] = [graph]
    return state_catalog


def buildAllEdges(dimensions, string=False, imaginary=False, loops=False):
    '''
    Given a collection of nodes, each with several possible colors/dimensions, returns all possible
    edges of the graph.
    
    Parameters
    ----------
    dimensions : list
        Accesible dimensions/colors for each of the nodes of the graph.
    string : boolean, optional
        If True, it returns a list of strings instead of tuples. 
    imaginary : boolean, str ('cartesian' or 'polar'), optional
        If False, it returns real weights.
    loops : boolean, optional
        Allow edges to connect twice the same node.
        
    Returns
    -------
    all_edges : list
        List of all possible edges given the dimensions of the nodes. 
        If string, it returns a list of strings.
        Else, it returns a list of tuples: (node1, node2, color1, color2).
    '''
    num_vertices = len(dimensions)
    all_edges = []
    if loops:
        combo_function = itertools.combinations_with_replacement
        for pair in combo_function(range(num_vertices), 2):
            if pair[0] == pair[1]:  # (node1, node1, 1, 0) is not stored
                for dims in combo_function(range(dimensions[pair[0]]), 2):
                    all_edges.append((pair[0], pair[1], dims[0], dims[1]))
            else:
                for dims in itertools.product(*[range(dimensions[ii]) for ii in pair]):
                    all_edges.append((pair[0], pair[1], dims[0], dims[1]))
    else:
        for pair in itertools.combinations(range(num_vertices), 2):
            for dims in itertools.product(*[range(dimensions[ii]) for ii in pair]):
                all_edges.append((pair[0], pair[1], dims[0], dims[1]))
    # returns edges whether as tuples or as sympy symbols
    if string:
        if imaginary == False:
            return [edgeWeight(edge) for edge in all_edges]
        else:
            return (['r_{}_{}_{}_{}'.format(*edge) for edge in all_edges] +
                    ['th_{}_{}_{}_{}'.format(*edge) for edge in all_edges])
    else:
        return all_edges


def stringEdges(edge_list, imaginary=False):
    '''
    Given a list of edges written as tuples it returns them in string format.
    
    Parameters
    ----------
    edge_list : list of tuples
        List of edges: [(node1, node2, color1, color2), ...]   
    imaginary : boolean, str ('cartesian' or 'polar'), optional
        If False, it returns real weights. Else, the weights are complex and defined by string pairs. 
        
    Parameters
    ----------
    list of str
        List of weight' strings.
    '''
    if imaginary == False:
        return ['w_{}_{}_{}_{}'.format(*edge) for edge in edge_list]
    else:
        return (['r_{}_{}_{}_{}'.format(*edge) for edge in edge_list]
                + ['th_{}_{}_{}_{}'.format(*edge) for edge in edge_list])


def buildRandomGraph(dimensions, num_edges, cover_all=True, loops=False):
    '''
    Given a set of nodes with different dimensions it build a random graph with a given number of edges.
    
    Parameters
    ----------
    dimensions : list
        Accesible dimensions/colors for each of the nodes of the graph.
    num_edges : int
        Number of edges of the random graph.
    cover_all : boolean, optional
        If True, it makes sure the selected edges cover all the nodes of the graph.   
        
    Returns
    -------
    graph : list of tuples
        List of randomly chosen edges: [(node1, node2, color1, color2), ...]
    '''
    all_edges = buildAllEdges(dimensions, loops=loops)
    if cover_all:
        num_nodes = len(dimensions)
        if 2 * num_edges >= num_nodes:
            all_covered = False
        else:
            raise ValueError('num_edges is too low to cover all nodes')
        while not all_covered:
            graph = random.sample(all_edges, num_edges)
            all_covered = (len(np.unique(np.array(graph)[:, :2])) == num_nodes)
        return sorted(graph)
    else:
        return sorted(random.sample(all_edges, num_edges))


def allPerfectMatchings(dimensions):
    '''
    Given a collection of nodes with different dimensions/colors available, it produces all possible
    states that erase from these and the different graphs that can produce such states.
    The graphs nodes can present different colors/dimensions and be connected by more than one edge.

    A fully connected (uncolored) graph with 2n nodes has (2n)!/((2**n)⋅n!) perfect matchings.

    Parameters
    ----------
    dimensions : list
        Accesible dimensions/colors for each of the nodes of the graph.
        
    Returns
    -------
    color_dict : dictionary
        Dictionary of available states. The keys are the different combinations of colored nodes
        (creator operators), that is, the state produced. 
        For each or state, the dictionary stores a list of all possible graphs that produce such state.
        The notation employed for the nodes is: (node, color/dimension).
        The notation employed for the edges is: (node1, node2, color1, color2).
        Node2 cannot be lower than Node1.
    '''
    num_nodes = len(dimensions)
    crowded_graph = [list(range(num_nodes))]
    color_dict = {}
    for crowd in crowded_graph:
        # For each list of nodes in crowded_graph (with possible repetitions) 
        # we store all dimensions/coloring the nodes can have in color_nodes.
        color_nodes = []
        # We distinguish between nodes but not between repeated nodes.
        # [(node1,color1),(node1,color2)] = [(node1,color2),(node1,color1)]
        for coloring in itertools.product(*[list(range(dimensions[nn])) for nn in crowd]):
            color_nodes.append(sorted([[crowd[ii], coloring[ii]] for ii in range(len(crowd))]))
        # After sorting the list of colored nodes, the next one erase duplicities.
        color_nodes = [[tuple(ed) for ed in graph] for graph in np.unique(color_nodes, axis=0)]
        for coloring in color_nodes:
            color_dict[tuple(coloring)] = allColorGraphs(coloring)
    return color_dict


def allEdgeCovers(dimensions, order=0, loops=False): # This function should replace allPerfectMatchings
    '''
    Given a collection of nodes with different dimensions/colors available, it produces all possible
    states that erase from these and the different graphs that can produce such states.
    The graphs nodes may have different colors/dimensions. The edges may be duplicated or form loops.

    Parameters
    ----------
    dimensions : list
        Accesible dimensions/colors for each of the nodes of the graph.
    order : int, optional
        Orders above the minimum required to cover all nodes.
        If 0, the output are only perfect machings, with all nodes having degree 1.
    loops : boolean, optional
        Allow edges to connect twice the same node. Only available for order>0.

    Returns
    -------
    color_dict : dictionary
        Dictionary of available states. The keys are the different combinations of colored nodes
        (creator operators), that is, the state produced.
        For each or state, the dictionary stores a list of all possible graphs that produce such state.
        The notation employed for the nodes is: (node, color/dimension).
        The notation employed for the edges is: (node1, node2, color1, color2).
        Node2 cannot be lower than Node1.
    '''
    num_nodes = len(dimensions)
    # Given a list with N different nodes, if order>0 some nodes will have degree
    # larger than 1, i.e., their creator operators will be repeated.
    # crowded_graph stores all ways in which these repeatitions may occur.
    # For N=4 and order=1: 000123, 001123, 001223, 001233, 011123, 011223...
    additions = list(itertools.combinations_with_replacement(range(num_nodes), 2 * order))
    crowded_graph = [list(range(num_nodes))] * len(additions)
    for ii, added in enumerate(additions):
        crowded_graph[ii] = sorted(crowded_graph[ii] + list(added))
    color_dict = {}
    for crowd in crowded_graph:
        # For each list of nodes in crowded_graph (with possible repetitions)
        # we store all dimensions/coloring the nodes can have in color_nodes.
        color_nodes = []
        # We distinguish between nodes but not between repeated nodes.
        # [(node1,color1),(node1,color2)] = [(node1,color2),(node1,color1)]
        for coloring in itertools.product(*[list(range(dimensions[nn])) for nn in crowd]):
            color_nodes.append(sorted([[crowd[ii], coloring[ii]] for ii in range(len(crowd))]))
        # After sorting the list of colored nodes, the next one erase duplicities.
        color_nodes = [[tuple(ed) for ed in graph] for graph in np.unique(color_nodes, axis=0)]
        for coloring in color_nodes:
            color_dict[tuple(coloring)] = allColorGraphs(coloring, loops)
    return color_dict


def recursivePerfectMatchings(graph, store, matches=[], edges_left=None):
    '''
    The heavy lifting of findPerfectMatchings.
    '''
    if edges_left == None: edges_left = len(np.unique(np.array(graph)[:, :2])) / 2
    if len(graph) > 0:
        for edge in targetEdges([nodeDegrees(graph)[0][0]], graph):
            recursivePerfectMatchings(removeNodes(edge[:2], graph), store,
                                      matches + [edge], edges_left - 1)
    elif len(graph) == 0 and edges_left == 0:
        store.append(sorted(matches))
    else:
        pass  # Some nodes are not matched and never will


def findPerfectMatchings(graph):
    '''
    Returns all possible perfect matchings (if any) of a graph.
    
    It starts by taking away the color/dimensions of the edges. Once the perfect matchings are build
    for the uncolored edges, it 'paint' them again with all possible combinations.
    
    Parameters
    ----------
    graph : list of tuples
        List of edges: [(node1, node2, color1, color2), ...].
        
    Returns
    -------
    painted_matchings : list of tuples
        Lists of all the possible perfect matchings of the input graph.
    '''
    avail_colors = edgeBleach(graph)
    raw_matchings = []
    recursivePerfectMatchings(list(avail_colors.keys()), raw_matchings)
    painted_matchings = []
    for match in raw_matchings:
        for coloring in itertools.product(*[avail_colors[edge] for edge in match]):
            color_match = [edge + color for edge, color in zip(match, coloring)]
            painted_matchings.append(tuple(color_match))
    return painted_matchings


def recursiveEdgeCover(graph, store, matches=[], edges_left=None, nodes_left=[], order=1, loops=False):
    '''
    Pseudorecursive function that append all possible edge covers of a
    graph in a given list 'store'.
    It does the heavy lifting of findEdgeCovers.

    Parameters
    ----------
    graph : list
        List of all available edges.
    store : list
        List to store the edge covers.
    matches : list, optional
        List of edges to build a perfect matching. If it fulfils all
        requirements it will end up in store
    edges_left : int, optional
        Total number of edges of each edge cover. If None, this is computed
        as the minimum edges required to cover all nodes plus the order.
    nodes_left : list, optional
        List of all nodes to be covered. By default, the list is obtained
        from the nodes covered in graph but it can be set to target
        specific nodes.
    order : int, optional
        Orders above the minimum required to cover all nodes. If 0, the
        output are only perfect machings, with all nodes having degree 1.
    loops : boolean, optional
        Allow edges to connect twice the same node. These cannot appear for
        perfect matchings.

    Returns
    -------
    None, but now the input list 'store' contains all possible edge covers.
    '''
    if len(nodes_left) == 0: nodes_left = np.unique(np.array(graph)[:, :2])
    if edges_left == None: edges_left = order + len(nodes_left) / 2
    case = 2 * edges_left - len(nodes_left)
    if case > 1:
        for edge in graph:
            recursiveEdgeCover(graph, store, matches + [edge], edges_left - 1,
                               [node for node in nodes_left if node not in edge[:2]], loops=loops)
    elif case == 1:
        for edge in targetEdges(nodes_left, graph):
            if edges_left > 1:
                new_nodes_left = [node for node in nodes_left if node not in edge[:2]]
                recursiveEdgeCover(targetEdges(new_nodes_left, graph), store,
                                   matches + [edge], edges_left - 1, new_nodes_left, loops=loops)
            if edges_left == 1 and not (sorted(matches + [edge]) in store):
                store.append(sorted(matches + [edge]))
    elif case == 0:  # We are in the perfect matching situation
        subgraph = [ed for ed in graph if (ed[0] in nodes_left) & (ed[1] in nodes_left)]
        if loops: subgraph = [edge for edge in subgraph if edge[0] != edge[1]]
        perfect_matchings = []
        recursivePerfectMatchings(subgraph, perfect_matchings, edges_left=edges_left)
        for pm in perfect_matchings:
            if not (sorted(matches + pm) in store):
                store.append(sorted(matches + pm))
    else:
        pass


# for large orders we could compute a first pack of edges at once using itertools


def findEdgeCovers(graph, edges_left=None, nodes_left=[], order=1, loops=False):
    '''
    Given a list of edges, returns all possible edge covers, up to a certain
    order or using a certain number of edges.

    Parameters
    ----------
    graph : list
        List of all available edges.
    edges_left : int, optional
        Total number of edges of each edge cover. If None, this is computed
        as the minimum edges required to cover all nodes plus the order.
    nodes_left : list, optional
        List of all nodes to be covered. By default, the list is obtained
        from the nodes covered in graph but it can be set to target
        specific nodes.
    order : int, optional
        Orders above the minimum required to cover all nodes. If 0, the
        output are only perfect machings, with all nodes having degree 1.
    loops : boolean, optional
        Allow edges to connect twice the same node. These cannot appear for
        perfect matchings.

    Returns
    -------
    covers : list
        List of graphs with all nodes having, at least, degree 1.
    '''
    avail_colors = edgeBleach(graph)
    raw_covers = []
    recursiveEdgeCover(list(avail_colors.keys()), raw_covers, edges_left=edges_left,
                       nodes_left=nodes_left, order=order, loops=loops)
    painted_covers = []
    for cover in raw_covers:
        for coloring in itertools.product(*[avail_colors[edge] for edge in cover]):
            color_cover = [edge + color for edge, color in zip(cover, coloring)]
            painted_covers.append(sorted(color_cover))
    return [[tuple(ed) for ed in graph] for graph in np.unique(painted_covers, axis=0)]


# # String Expressions
# Write the string expressions used in the optimization.

def edgeWeight(edge, imaginary=False):
    '''
    It writes the edge's weight as needed for `writeNorm`, `targetEquation`, and others.
    
    Parameters
    ----------
    edge : tuple
        Edge in tuple notation: (node1, node2, color1, color2).
    imaginary : boolean, str ('cartesian' or 'polar'), optional
        If False, the final expression uses only real weights. If 'cartesian' or 'polar', the weights 
        are complex values written in the corresponding notation.
        
    Returns
    -------
    str
        Edge's weight written as string for contributions on a particular ket.
    '''
    if imaginary == False:
        return 'w_{}_{}_{}_{}'.format(*edge)
    elif imaginary == 'cartesian':
        return '(r_{}_{}_{}_{}+1j*th_{}_{}_{}_{})'.format(*edge * 2)
    elif imaginary == 'polar':
        return 'r_{}_{}_{}_{}*np.exp(1j*th_{}_{}_{}_{})'.format(*edge * 2)
    else:
        raise ValueError('Introduce a valid input `imaginary`.')


def weightProduct(edge_list, imaginary=False):
    '''
    It writes the product of edge's weights as needed for `writeNorm`, `targetEquation`, and others.
    
    Parameters
    ----------
    edge_list : list of tuples
        List of edges: [(node1, node2, color1, color2), ...].
    imaginary : boolean, str ('cartesian' or 'polar'), optional
        If False, the final expression uses only real weights. If 'cartesian' or 'polar', the weights 
        are complex values written in the corresponding notation.
        
    Returns
    -------
    str
        Product of edge's weights written as string for contributions on a particular ket.
    '''
    return '*'.join([edgeWeight(edge, imaginary) for edge in edge_list])


# This expression could also be UNlambdified so we add some check for the Counter
factProduct = lambda lst: np.product([factorial(ii) for ii in Counter(lst).values()])

def writeNorm(state_catalog, imaginary=False):
    '''
    Write a normalization function with all the states of a dictionary.
    
    Parameters
    ----------
    state_catalog : dictionary
        Dictionary with all the kets created by each graph as keys. If more than one graph leads to
        the same state, they are listed together in the corresponding entrance of the dictionary.
    
    imaginary : boolean, str ('cartesian' or 'polar'), optional
        If False, the final expression uses only real weights. If 'cartesian' or 'polar', the weights 
        are complex values written in the corresponding notation.
        
    Returns
    -------
    str
        Norm string with the contributions of each perfect matching to their corresponding ket.
    '''
    norm_sum = []
    for key, values in state_catalog.items():
        # The division with factProduct comes from combinatoric reasons: wikipedia.org/wiki/Multinomial_theorem
        term_sum = [f'{weightProduct(graph, imaginary)}/{factProduct(graph)}' for graph in values]
        term_sum = ' + '.join(term_sum)
        if imaginary == False:
            # The multiplication with factProduct from creator operators: wikipedia.org/wiki/Creation_operators
            # The factor should be squared or, in this case, left outside the square
            norm_sum.append(f'{factProduct(key)}*(({term_sum})**2)')
        else:
            norm_sum.append(f'{factProduct(key)}*(abs({term_sum})**2)')
    return ' + '.join(norm_sum).replace('/1 +', ' +').replace('/1)', ')').replace('+ 1*', '+ ')


def targetEquation(ket_list, amplitudes=None, state_catalog=None, imaginary=False):
    '''
    Introducing the amplitudes for each ket, it writes a non-normalized fidelity function with all 
    the ways the state can be build stored in state_catalog.
    
    If no state_catalog is introduced it builds all possible graphs that generate the desired kets.
    
    Parameters
    ----------
    ket_list : list of tuples
        List of kets written as: [((node1, dimension1), (node2, dimension2), ...), ...].
        
    amplitudes : list, optional
        List of the amplitudes we aim for each ket. If None, all will be the same.
        
    state_catalog : dictionary, optional
        Dictionary with all the kets created by each graph as keys. If more than one graph leads to
        the same state, they are listed together in the corresponding entrance of the dictionary.
    
    imaginary : boolean, str ('cartesian' or 'polar'), optional
        If False, the final expression uses only real weights. If 'cartesian' or 'polar', the weights 
        are complex values written in the corresponding notation.
        
    Returns
    -------
    str
        Non-normalized fidelity string with all the contributing weights.
    '''
    if amplitudes == None:
        amplitudes = [1] * len(ket_list)
    else:
        if len(amplitudes) != len(ket_list):
            raise ValueError('The number of amplitudes and states should be the same')
    if imaginary == 'polar':
        amplitudes = [amp[0] * np.exp(1j * amp[1]) for amp in amplitudes]
    norm2 = abs(np.conjugate(amplitudes) @ amplitudes)
    if norm2 != 1: # Is this useless? I think so (Carlos)
        norm2 = str(norm2)
    if state_catalog == None:
        state_catalog = {tuple(ket): allColorGraphs(ket) for ket in ket_list}
    equation_sum = []
    for coef, ket in zip(np.conjugate(amplitudes), ket_list):
        # The division with factProduct comes from combinatoric reasons: wikipedia.org/wiki/Multinomial_theorem
        term_sum = [f'{weightProduct(graph, imaginary)}/{factProduct(graph)}' for graph in state_catalog[tuple(ket)]]
        term_sum = ' + '.join(term_sum)
        # The multiplication with factProduct from creator operators: wikipedia.org/wiki/Creation_operators
        creation_term = factProduct(ket)
        if (creation_term**.5) % 1 == 0:
            creation_term = int(creation_term**.5)
        else:
            creation_term = f'{creation_term}**.5'
        equation_sum.append(f'({coef})*({creation_term})*({term_sum})')
    equation_sum = ' + '.join(equation_sum)
    if imaginary == False:
        return f'(({equation_sum})**2)/{norm2}'.replace('/1 +', ' +').replace('/1)', ')').replace('(1)*', '')
    else:
        return f'(abs({equation_sum})**2)/{norm2}'.replace('/1 +', ' +').replace('/1)', ')').replace('(1)*', '')


def buildLossString(loss_function, variables):
    loss_string = 'lambda ' + ', '.join(variables) + f': {loss_function}'
    loss_string = f'func = lambda inputs: ({loss_string})(*inputs) '
    exec(loss_string, globals())
    return func, loss_string  # we can keep the second as a security check


# # String Expressions
# Write the string expressions used in the optimization.

def ptrace(u, keep, dims, optimize=False):
    """Calculate the partial trace of an outer product

    ρ_a = Tr_b(|u><u|)

    From slek120: https://scicomp.stackexchange.com/a/30057
    
    Parameters
    ----------
    u : array
        Vector to use for outer product
    keep : array
        An array of indices of the spaces to keep after being traced. For instance, if the space is
        A x B x C x D and we want to trace out B and D, keep = [0,2]
    dims : array
        An array of the dimensions of each space. 
        For instance, if the space is A x B x C x D, dims = [dim_A, dim_B, dim_C, dim_D]

    Returns
    -------
    ρ_a : 2D array
        Traced matrix
    """
    keep = np.asarray(keep)
    dims = np.asarray(dims)
    Ndim = dims.size
    Nkeep = np.prod(dims[keep])

    idx1 = [i for i in range(Ndim)]
    idx2 = [Ndim + i if i in keep else i for i in range(Ndim)]
    u = u.reshape(dims)
    rho_a = np.einsum(u, idx1, u.conj(), idx2, optimize=optimize)
    return rho_a.reshape(Nkeep, Nkeep)


def compute_entanglement(qstate: np.array, sys_dict: dict, var_factor=0) -> float:
    """
    calculate for a set of bipartions given in config the mean of trace[ rho_A ], where rho_A is 
    reduced density matrix of given state for the given bipartitions

    Parameters
    ----------
    qstate : np.array
        basis vector of corrosponding state as np.array
    sys_dict : dict
        that stores essential_infos (see help_functions)

    Returns
    -------
    float
        return sum_bipar[ trace( rho_bi ** 2)  ]

    """
    dimi = np.array(sys_dict['dimensions'])
    try:  # for checking if norm is not zero -> if so return 1 cause no ket
        qstate *= 1 / (np.linalg.norm(qstate))
    except TypeError:
        return 1

    def calc_con(mat, par):
        red = ptrace(mat, par, dimi, False)
        return np.einsum('ij,ji', red, red)  # is equivalent to trace( red**2 ) but faster

    loss_vec = [calc_con(qstate, par[0]) for par in sys_dict['bipar_for_opti']]
    lenght = len(loss_vec)
    mean = sum(loss_vec) / lenght
    if var_factor == 0:  # no need to compute variance if factor = 0
        return mean
    else:
        var = sum([(x - mean) ** 2 for x in loss_vec]) / (lenght)
        return mean + var_factor * var


def entanglement_fast(avail_states: dict, sys_dict: dict,var_factor):
    """
    compute the entanglement according to compute_entanglement() of the state given by the graph
    according to given avail_states

    Parameters
    ----------
    sys_dict : dict
        that stores essential infos of quantum system (see helpfunctions.get_sysdict).
    avail_states : dict
        storing graphs(value) for each state(key)

    Returns
    -------
    str
        returns entanglement of graph only in terms of graphs weights.

    """

    state_vector = sys_dict['dim_total'] * [0]
    num_anc = sys_dict['num_ancillas']  # count ancillas
    if num_anc != 0:
        temp_dic = dict()
        for key, val in avail_states.items():
            try:
                extsting_graph = temp_dic[key[:-num_anc]]
                extsting_graph.append(val[0])
                temp_dic[key[:-num_anc]] = extsting_graph 
            except KeyError:
                temp_dic[key[:-num_anc]] = val
        avail_states = temp_dic
    for idx, state in enumerate(sys_dict["all_states"]):
        try:
            term_sum = [weightProduct(graph, sys_dict['imaginary'])
                        for graph in avail_states[state]]
            term_sum = '+'.join(term_sum)
            state_vector[idx] = f'{(term_sum)}'
        except KeyError:
            pass

    if sys_dict['imaginary'] is False:
        return f'compute_entanglement(np.array({state_vector} ),'.replace("'", "") \
               + f' {sys_dict}, {var_factor} )'
    else:
        return f'abs(compute_entanglement(np.array({state_vector} ),'.replace("'", "") \
               + f' {sys_dict}, {var_factor} ))'



# -*- coding: utf-8 -*-
"""
Created on Sat Jul 30 10:51:51 2022

@author: janpe
"""
import os

from cmath import polar
from pytheus.fancy_classes import State, Graph
from pathlib import Path
import json
import matplotlib.pyplot as plt
import numpy as np
import pytheus.graphplot as gp
import pytheus.help_functions as hf
from scipy.linalg import logm
from pytheus.theseus import ptrace


def convert_graph_keys_in_tuple(graph: dict) -> dict:
    """
    here we can convert our Graph dict in a dict that has strings as keys
    we need this to save it in json file
    """
    # convert keys in str
    ret_dict = {}
    for key in graph.keys():
        if type(key) is str:
            try:
                ret_dict[eval(key)] = graph[key]
            except:
                pass

    return ret_dict


def convert_file_path_to_dic(abs_file_path: Path) -> dict:
    with abs_file_path.open() as file:
        dictionary = json.load(file)
    file.close()
    return dictionary


def string_wrapper(string: str, max_chars_per_line=80, 
                   latex = '$',
                   line_break_char = '+' ) -> str:
    """
    wraps a string to a given line lenght 

    Parameters
    ----------
    string : str
        string has to be wrapped.
    max_chars_per_line : TYPE, optional
        max nums of char in one line. The default is 80.

    Returns
    -------
    str
        wrapped string

    """
    new_string = latex
    counts = 0
    for ss in string:
        if counts >= max_chars_per_line and ss == line_break_char:
            new_string += ss + latex + '\n' + latex
            counts = 0
        else:
            new_string += ss
            counts += 1

    new_string += latex
    return new_string


def num_in_str(num, dec=2, change_sign=False) -> str:
    """
    make sure to convert a number in a proper Latex string
    output depending if it is complexe or not

    Parameters
    ----------
    num : TYPE
        number to convert
    dec : TYPE, optional
        dec for rounding. The default is 2.
    change_sign : TYPE, optional
        needed in state analyzer, changes sign of num The default is False.

    Returns
    -------
    Str
        returns num in Latex string format

    """
    if np.round(num.imag, dec) != 0:
        polar_num = polar(num)
        if not change_sign:
            num = np.round([polar_num[0], polar_num[1]/np.pi],  dec)
        else:
            num = np.round([-polar_num[0], polar_num[1]/np.pi],  dec)
        return '{} \cdot e^{{ {} \cdot i \cdot \pi }}'.format(*num).replace('1.0 * ', '')
    else:
        rounded = np.round(num.real, dec)
        if not change_sign:
            if rounded == 1 or rounded == -1:
                return ''
            return f'{np.round(num.real,dec)}'
        else:
            if rounded == 1 or rounded == -1:
                return ''
            return f'{-np.round(num.real,dec)}'


entanglement_functions = {
    'con': lambda matrix: abs((2*(1 - min(np.trace(matrix@matrix), 1)))**0.5),
    'entropy': lambda matrix: -np.trace(matrix@logm(matrix)),
    'dense': lambda matrix: np.trace(matrix@matrix)}



class entanglement_measure():

    def __init__(self, dimensions: list, measure_kind='con'):
        """
        a class to calculate a maximal values for a given measure kind
        of entanglment:
            - all reduced density being maximal mixed 
            - best values for a given bipartion
            - the sum of best values for a given k-bipartion
            - all maximal values
        Parameters
        ----------
        dimensions : list
            Ddimension list of system
        measure_kind : func, optional
            measure function according to dict above . The default is 'con'.

        Returns
        -------
        None.

        """
        if dimensions.count(1) == 0:
            self.dims = dimensions
        else:
            self.dims = dimensions[:-dimensions.count(1)]
        self.num_par = len(self.dims)
        self.densitys = self.maximal_mixed_densitys()
        self.func = entanglement_functions[measure_kind]
        self.m = measure_kind
        self.len = self.amount_bipar()

    def maximal_mixed_densitys(self) -> list:
        """
        return a list containing all reduced density matrices for that
        are maximally mixed:
            e.g: for dim = [2,2,2,2] we have 2 splittings:
                - one particle and the rest
                - and 2 particles and 2 particles
        -> returns list  maximal mixed density matrix for those two splits
        """
        densitys = []
        for kk in range(1, int(self.num_par/2)+1):
            dim_subsystem = np.prod(self.dims[:kk])
            matrix = 1/dim_subsystem * np.identity(dim_subsystem)
            densitys.append(matrix)
        return densitys

    def max_value_given_len_bipar(self, k: int):
        red_density = self.densitys[k-1]
        return self.func(red_density)

    def max_all(self, k_uniform=0) -> float:
        """
        calculate the maximal sum of given entanglement measurement 

        k_uniform : int, optional
            which bipartion of lenght k one consider,
            if k = 0:
            the sum of all possible bipartion and the corrosponding
            max value will be returned

        """

        if k_uniform == 0:
            return sum([self.max_value_given_len_bipar(min(len(bipar[0]), len(bipar[1])))
                        for bipar in
                        hf.get_all_bi_partions(self.num_par)])

        else:
            return sum([self.max_value_given_len_bipar(len(bipar[0])) for bipar in
                        hf.get_all_bi_partions(self.num_par)
                        if len(bipar[0]) == k_uniform])

    def amount_bipar(self):
        """
        calculate amount bipars for all ks
        eg:    dim = [2,2,2,2] :
            [4,3] we have 4 splits with 3|1 particles
            and 3 splits with 2|2 particles

        """
        lenghts = []
        for kk in range(1, 1+int(self.num_par/2)):
            ls = [1 for bipar in hf.get_all_bi_partions(self.num_par)
                  if len(bipar[0]) == kk]
            lenghts.append(len(ls))
        return lenghts

    def max_values(self,  k_uniform=0, full_vec=False) -> list:
        """
        return vector containing all maximal values for all possible
        bipartions (when k_uniform=0) else returns a vector with all
        possible bipartions having len(k_uniform), 
        if  full_vec = True: returns vec for all bipartions
        """

        if k_uniform == 0:
            max_val_all_bipar = [self.max_value_given_len_bipar(len(bipar[0]))
                                 for bipar in
                                 hf.get_all_bi_partions(self.num_par)]
            if full_vec:
                return np.array(max_val_all_bipar)
            else:
                return np.unique(max_val_all_bipar)

        else:
            max_val_k_bipar = [self.max_value_given_len_bipar(len(bipar[0]))
                               for bipar in
                               hf.get_all_bi_partions(self.num_par)
                               if len(bipar[0]) == k_uniform]
            if full_vec:
                return np.array(max_val_k_bipar)
            else:
                return np.unique(max_val_k_bipar)[0]

    def info(self, returne=False):
        string = f'For {self.num_par} particles with dimension {self.dims[0]}:\n'
        string += f'max {self.m} : {self.max_all():.3f}\n'
        for kk in range(1, 1+len(self.densitys)):
            max_val = self.max_value_given_len_bipar(kk)
            string += f'For k = {kk}: {max_val:.3f} ( {self.len[kk-1]} )  \n'

        if returne:
            return string
        else:
            print(string)


class state_dic_class(dict):
    """
    class to make sure that we can use both inputs:
        dic['001'] = x or
        dic[ [(0,0),(1,0),(2,1)] ] = x
    """

    def __init__(self, dic=None):
        super().__init__()
        if dic is not None:
            for kk, vv in dic.items():
                self.__setitem__(kk, vv)

    def __getitem__(self, keys):
        if isinstance(keys, str):
            return super().__getitem__(keys)
        if isinstance(keys, list):
            return super().__getitem__(hf.stateToString(keys))
        else:
            raise ValueError('only keystrings or state list allowed as keys')

    def __setitem__(self, key, item):
        if isinstance(key, str):
            super().__setitem__(key, item)
        else:
            super().__setitem__(hf.stateToString([key]), item)

    def items(self, as_string=True):
        if as_string:
            return super().items()
        else:
            ket = [hf.makeState(kk) for kk in super().keys()]
            return zip(ket, super().values())


class state_analyzer():

    def __init__(self, state: State,  weights=[], dim=[], precision=1e-3,
                 measure_kind='entropy'):
        self.dic = state_dic_class(state.state)
        self.norm = state.norm
        self.pre = precision
        self.len = len(self.dic)  # number of summands
        if len(weights) != 0:
            assert self.len == len(weights), "lenght of weight should match"

        if len(dim) == 0:
            ni = len(list(self.dic.keys())[0])
            self.dim = ni * [1 + max([int(x) for ket in
                                      self.dic.keys() for x in ket])]
        else:
            self.dim = dim
        self.check_ancillas()
        self.num_par = len(self.dim)   # number particles
        self.func = entanglement_functions[measure_kind]
        self.allbipar = list(hf.get_all_bi_partions(self.num_par))
        self.max = entanglement_measure(self.dim, measure_kind=measure_kind)
        self.measure_kind = measure_kind
    def check_ancillas(self):
        """
        check if given dic has more particles than dim (means we have ancilla)
        (e.g dim = [2,2,1]) -> reduce to dim=[2,2] because 1 <-> ancilla

        Returns
        -------
        None.

        """

        num_ancillas = self.dim.count(1)
        if num_ancillas != 0:
            temp_dic = dict()
            for key, val in self.dic.items():
                temp_dic[key[:-num_ancillas]
                         ] = temp_dic.setdefault(key[:-num_ancillas], 0) + val
            self.dic = temp_dic
            self.dim = self.dim[:-num_ancillas]

    def state_vec(self, normalized=False) -> np.array:
        """
        calculate the lineare algebra vector for the given state.
        e.g. |00> + |10> --> [ [1],[0],[1],[0] ] 

        Parameters
        ----------
        normalized : bool, optional
            if state vec should be normalized. The default is False.

        Raises
        ------
        TypeError
            when state vec is zero vector

        Returns
        -------
        np.array
            state vector

        """
        if any(isinstance(ampl, complex) for ampl in self.dic.values()):
            state_vec = np.zeros(np.product(self.dim), dtype=np.complex64)
        else:
            state_vec = np.zeros(np.product(self.dim))
        for idx, ket in enumerate(hf.get_all_kets_for_given_dim(self.dim, str)):
            try:
                state_vec[idx] = self.dic[ket]
            except KeyError:
                pass
        if all(state_vec == 0) is True:
            raise TypeError('State Vector is zero for all entrys, hmm?')
        self.norm = np.linalg.norm(state_vec)
        if normalized:
            return state_vec * 1/(self.norm)
        else:
            return state_vec

    def ent(self):
        """
        calculate total sum over all bipartions for given ent function 
        in self.func

        Returns
        -------
        float
            sum of ent measure for all bipartions.

        """
        qstate = self.state_vec(normalized=True)

        def calc_ent(mat, par):
            red = ptrace(mat, par, self.dim)
            return self.func(red)
        self.con_vec = np.array([calc_ent(qstate, par[0]).real for par in
                                 self.allbipar])

        self.c = sum(self.con_vec)

        def normalizer(x, min_val):
            # little function to make sure that we normalize values
            # for dense : x = min_val -> 1 and x = 1 -> 0
            # for concurrence or entropy just diveded by highest poss. value
            if self.measure_kind == 'dense':
                alpha = 1/(min_val-1)
                return alpha * x - alpha
            else:
                return x/min_val
        
        self.con_vec_norm = np.array([normalizer(calc, best) for calc, best in zip(self.con_vec,
                                                                                   self.max.max_values(full_vec=True))])

        return self.c

    def get_reduced_density(self, bipar, normalized=False) -> np.array:
        """

        Parameters
        ----------
        bipar : list/tuple
            of given bipartion, e.g [0,1] for split AB|CD
        normalized : TYPE, optional
            if normalizing state vector The default is False.

        Returns
        -------
        np.array
            reduced density matrix.

        """
        return ptrace(self.state_vec(normalized), bipar, self.dim)

    def print_red_densitys(self, k):
        self.calc_k_uniform()  # for getting  k mask
        for bipar in np.array(self.allbipar)[self.k_mask[k-1]]:
            print(f'{bipar} :')
            print(self.get_reduced_density(bipar[0], True))

    def state_string(self, dec=2, filter_zeros=False, with_color=False):
        """
        get a string of the state in Latex format

        Parameters
        ----------
        dec : int, optional
            for decimal roundings of amplitudes. The default is 2.
        filter_zeros : bool, optional
            if one displays the kets with amplitude = 0 after rounding.
            The default is False.
        with_color : bool, optional
            DESCRIPTION. The default is False.

        Returns
        -------
        string_end : str
            Latex string of state

        """
        if with_color:
            def st_col(strg, col): return r' \color{{{0}}} '.format(col) + strg
        else:
            def st_col(strg, col): return strg

        ampls = [np.round(amp/self.norm, dec) for amp in self.dic.values()]
        most = np.round(
            abs(max(set([polar(aa)[0] for aa in ampls]), key=ampls.count)), dec)
        if most < 0.001:
            most = 1
        most = 1
        strs_plus, strs_min = [], []
        cet_counts = 0
        for ket, ampl in sorted(self.dic.items()):
            ampl = np.round(ampl/self.norm, dec)
            if ampl.real > 0 or ampl.imag != 0:
                ampl = np.round(1/most * ampl, dec)
                if abs(ampl) < float('10e-{}'.format(dec)):
                    col = 'red'
                else:
                    col = 'black'
                if abs(ampl) != 0 or not filter_zeros:
                    strs_plus.append(
                        st_col(f' + {num_in_str(ampl)} \cdot', col))
                    strs_plus.append(r'|{0} \rangle'.format(ket))
                    cet_counts += 1
            else:
                ampl = np.round(1/most * ampl, dec)
                if abs(ampl) < float('10e-{}'.format(dec)):
                    col = 'red'
                else:
                    col = 'black'
                if abs(ampl) != 0 or not filter_zeros:
                    strs_min.append(
                        st_col(f' + {num_in_str(ampl,change_sign=True)} \cdot', col))
                    strs_min.append(r'| {0} \rangle'.format(ket))
                    cet_counts += 1
        try:
            strs_plus[0] = strs_plus[0].replace('+', '')
        except IndexError:
            pass

        strs_plus.append(st_col(' - (', 'black'))
        strs_plus += strs_min
        strs_plus.append(st_col(') ', 'black'))

        string_end = "".join(strs_plus).replace('- ()', '')
        string_end = string_wrapper(string_end[1:])

        return string_end

    def calc_k_uniform(self) -> (int, np.array):
        """
        get k uniform of state and normalized vector for each k-bipartion
        where 1 means  maximal mixed and 0 seperable for that bipartion
        e.g. :
            return is ( 1 , [ [1,1,1,1], [0,1,1] ] ) means state is 1-uniform
            and for all bipartions with cardinality 2 also maximal entangeld
            except for the first split AB|CD (= 0) it is separable

        Returns
        -------
        string_end : (int, np.array)
            num of k-uniform for given state and 
            normalized array for all bipartions

        """

        try:
            self.con_vec_norm
        except AttributeError:
            self.ent()

        # amount bipars for each k (as list)
        self.k_mask = []
        for kk in range(int(self.num_par/2)):
            self.k_mask.append([idx for idx, bi in enumerate(self.allbipar)
                                if min(len(bi[0]), len(bi[1])) == kk + 1])
        k = 0
        k_levels = []
        for kk in range(int(self.num_par/2)):
            bipar_kk = self.con_vec_norm[self.k_mask[kk]]
            # calc mean for each k for return
            k_levels.append(np.mean(bipar_kk))
            if all(np.isclose(bipar_kk, 1, rtol=self.pre)) is True:
                k += 1
        return k, k_levels

    def info_string(self, *args, **kwargs):
        """
        returns a string with all infos specified in args and kwargs

        Parameters
        ----------
        *args : TYPE
            DESCRIPTION.
        **kwargs : TYPE
            DESCRIPTION.

        Returns
        -------
        info_string : str
            info string

        """
        options = {
            'filter_zeros': False,
            'dec': 2,
            'with_color': False}
        if options['with_color']:
            def st_col(strg, col): return r'\color{0}'.format(col) + strg
        else:
            def st_col(strg, col): return strg

        if len(args) == 0:
            args = ['norm']
        options.update(kwargs)
        self.ent()
        info_string = ''
        if any(x in args for x in ['k', 'K', 'k-uniform']):

            k, k_level = self.calc_k_uniform()
            info_string += f'\nk-uniform: {k}'
            for ix, klev in enumerate(zip(k_level, self.max.len)):
                if np.isclose(klev[0], 1, rtol=self.pre):
                    col = 'green'
                else:
                    col = 'red'
                info_string += st_col(
                    f'\nk={ix + 1}: mean = {klev[0]:.3f} ({klev[1]})', col)

        if any(x in args for x in ['concurrence', 'ent']):
            info_string += f'\ntotal {self.measure_kind}: {round(self.c,options["dec"])}'
            info_string += f'\n {self.measure_kind} vector:'
            for kk, mask in enumerate(self.k_mask):
                k_info = f'\n k = {kk+1} : {[round(ii,options["dec"]) for ii in self.con_vec_norm[mask]]}'
    
                info_string += string_wrapper(k_info,55,line_break_char = ',',latex='',)
        if any(x in args for x in ['n', 'norm']):
            info_string += f'\n normalized by: 1/{round(self.norm,options["dec"])}'

        return info_string


class analyser():

    def __init__(self, folder, only_pm=False):
        if isinstance(folder, str):
            self.folder = Path(folder)
        else:
            self.folder = folder
        self.check_folder_name()
        self.get_all_states_in_folder_and_summary_file()
        self.dim = self.summary['dimensions']
        self.imaginary = self.summary['imaginary']
        self.only_pm = only_pm

    def check_folder_name(self):
        if len(list(Path(self.folder).rglob('*.json'))) == 0:
            raise ValueError(
                f'The given path {self.folder} does not contain jsons')
        elif self.folder.exists() is False:
            raise ValueError(f'The given path {self.folder} does not exist')

    def get_lossfunc_names(self):
        los = self.summary["loss_func"]
        if los == 'ent':
            k_unif = str(self.summary["K"])
            return [r'$\mathcal{{L}}_{{ K ={0} }}$'.format(k_unif)]
        elif los == 'cr':
            thres = self.summary["thresholds"]
            return [r'$1 - Countrate_{{ THR = {0} }}$'.format(thres[0]),
                    r'$1 - Fidelity_{{ THR ={0} }}$'.format(thres[1])]
        elif los == 'fid':
            thres = self.summary["thresholds"]
            return [r'$1 - Fidelity_{{ THR = {0} }}$'.format(thres[0]),
                    r'$1 - Countrate_{{ THR = {0} }}$'.format(thres[1])]
        else:
            return [f'loss {ii}' for ii in range(len(self.best['history'][0]))]

    def get_all_states_in_folder_and_summary_file(self):
        # iterate through all files in folder ending with json and append to files
        self.files = []
        for path in Path(self.folder).rglob('*.json'):
            if 'summary' in str(path):
                self.summary = convert_file_path_to_dic(path)
            elif 'best' in str(path):
                self.best = convert_file_path_to_dic(path)
            else:
                dic = convert_file_path_to_dic(path)
                dic['file_name'] = str(path.name)
                self.files.append(dic)

        self.files = sorted(self.files, key=lambda dic: dic["loss"])
        for idx, file in enumerate(self.files):
            print(f'({idx}): {file["file_name"]}')

        self.loss_func_names = self.get_lossfunc_names()

    def plot_losses(self):
        """
        plots y-axis: concurrence, x-axis: number of edges for a whole
        list of state dicts, plotvline plots a vertical line for a given positon

        Parameters
        ----------
        sysdic_list : list
            of state dicts
        plotvline : float, optional
            Position of vertical line. The default is 0.
        plot_mean : Boolean, optional
            IF mean should be plotted. The default is True.

        """

        def plotter(axe, num_loss):
            NUM_COLORS = len(self.files)
            cm = plt.get_cmap('gist_rainbow')
            axe.set_prop_cycle(color=[cm(1.*i/NUM_COLORS)
                                      for i in range(NUM_COLORS)])

            for idx, st in enumerate(self.files):
                loss_history = st['history']
                min_edge = len(st['graph'])
                edges = [ii for ii in range(
                    min_edge, min_edge+len(loss_history))]
                edges.reverse()
                loss = [ll[num_loss] for ll in loss_history]
                axe.plot(edges, loss, label=idx, alpha=0.4, lw=3)
                axe.plot(edges, loss, alpha=0.6, lw=0.7, color="black")
            axe.set_xlabel("Amount of edges left")
            axe.set_ylabel(self.loss_func_names[num_loss])
            axe.invert_xaxis()
            axe.legend()
            axe.grid()
        fig, axs = plt.subplots(len(self.best['loss']), 1)
        try:
            for idx_loss, ax in enumerate(axs):
                plotter(ax, idx_loss)
        except TypeError:
            plotter(axs, 0)

        plt.tight_layout()
        plt.show(block=False)
   

    def turn_dic_in_graph_state(self, graph_dict: dict, thresholds_amplitudes=np.inf):
        """
        returns a GRaph object and a state_analyzer object for given state

        Parameters
        ----------
        graph_dict : dict
            dict representing a graph.
        thresholds_amplitudes : TYPE, optional
            DESCRIPTION. The default is np.inf.

        Returns
        -------
        TYPE
            DESCRIPTION.

        """
        if self.only_pm:
            for ket, ampl in graph_dict.items():
                if self.summary['imaginary'] is False:
                    graph_dict[ket] = 1 if ampl > 0 else -1
                else:
                    graph_dict[ket] =  [1,ampl[1]] if ampl[0] > 0 else [-1,ampl[1]] 

        graph = Graph(graph_dict, dimensions=self.dim,
                      imaginary=self.imaginary)
        if self.imaginary is not False:
            graph.toCartesian()
        graph.getState()

        return graph, state_analyzer(graph.state, dim=self.dim)

    def all_perfect_matchings_to_pdf(self, nth_file=0, other_weights=[],
                                                   given_ket_only="", show=True, row_len=True,
                                                   dpi=100) -> None:
        """


        Parameters
        ----------
        state_sys : dict
            DESCRIPTION.
        other_weights : TYPE, optional
            DESCRIPTION. The default is [].
        given_ket_only : TYPE, optional
            DESCRIPTION. The default is "".
        show : TYPE, optional
            DESCRIPTION. The default is True.
        row_len : TYPE, optional
            DESCRIPTION. The default is True.

        Raises
        ------
        ValueError
            DESCRIPTION.

        Returns
        -------
        None
            DESCRIPTION.

        """

        state_sys = self.files[nth_file]
        pt = self.folder
        pt = pt / 'state_pdfs' / 'tmp'  # move data directory
        pt.mkdir(parents=True, exist_ok=True)
        savepath = str(pt)
        for f in os.listdir(savepath):
            os.remove(os.path.join(savepath, f))

        graph, __ = self.turn_dic_in_graph_state(state_sys['graph'])

        graph.getStateCatalog()
        plt.ioff()
        cat = graph.state_catalog
        if len(given_ket_only) != 0:
            if type(given_ket_only) == str:
                given_ket_only = [given_ket_only]
            for ket_i in given_ket_only:
                try:
                    th_state = hf.stringToTerm(ket_i)
                    cat = {th_state: cat[th_state]}
                except KeyError:
                    raise ValueError(
                        "The given Ket must be wrong, Graph does not have it")
        # calculation how many rows
        if row_len:
            row_len = max([len(pm_graphs) for pm_graphs in cat.values()])

        # calculation for coloring background
        colors = []
        for kk, vv in cat.items():
            total_weight = 0
            for ii, cover in enumerate(vv):
                weights_for_cover = [graph[edge] for edge in cover]
                total_weight += np.prod(weights_for_cover)

            if round(total_weight.real, 2) == 0:
                colors.append((1, 0.89, 0.77))
            else:
                colors.append((1, 1, 1))
        num_kets = len(self.dim) - self.dim.count(1)  # eliminate ancillas
        # determinate fontsize:
        fontsize = max(20, 33-num_kets-row_len)
        amount_pm = len(cat.items())
        for idx, (kk, vv) in enumerate(cat.items()):
            ket_string = "".join([str(int(k[1])) for k in kk[:num_kets]])
            total_weight = 0
            figgy, ax = plt.subplots(1, row_len,
                                     figsize=(row_len*800/dpi, 800/dpi),
                                     dpi=dpi)
            
            if row_len == 1:
                ax = [ax]  # make ax subscriptable when lenght one
            for ii, cover in enumerate(vv):
                weights_for_cover = [graph[edge] for edge in cover]
                total_weight += np.prod(weights_for_cover)
                figgy = gp.graphPlot(Graph(cover, weights=weights_for_cover),
                                     show=False, weight_product=True,
                                     show_value_for_each_edge=True,
                                     ax_fig=(figgy, ax[ii]), fontsize=fontsize
                                     )

            title = r'$  {0} \ket{{ {1} }} $'.format(
                    num_in_str(total_weight), ket_string)

            props = dict(boxstyle='round', facecolor='lightgrey')
            figgy.suptitle(title, y=0.99, fontsize=35, bbox=props)
            figgy.patch.set_facecolor(colors[idx])

            for ii in range(len(ax)):
                ax[ii].axis('off')
            plt.tight_layout()
            figgy.savefig(os.path.join(savepath, ket_string + '(' + str(ii) +
                                       ')' + ".jpeg"),
                          dpi=dpi, bbox_inches='tight',
                          facecolor=figgy.get_facecolor(), edgecolor='none')
            x = int(10*idx/amount_pm)
            print("state {}: [{}{}] {}/{}".format(nth_file,"#"*x, "."*(10-x), idx, amount_pm),
                  end='\r', flush=True)
            plt.close(figgy)

        self.convert_img_to_pdf(savepath, 1, self.files[nth_file]['file_name'])
        plt.ion()

    def convert_img_to_pdf(self, savepath, row_len, file_name='pm_pdf', show=False):
        from PIL import Image

        images = [
            Image.open(os.path.join(savepath, f))
            for f in os.listdir(savepath)
        ]

        def pil_grid(images, max_horiz=np.iinfo(int).max):
            def get_grid(func):
                n_images = len(images)
                n_horiz = func(n_images, max_horiz)
                h_sizes, v_sizes = [0] * n_horiz, [0] * (n_images // n_horiz)

                for i, im in enumerate(images):
                    h, v = i % n_horiz, i // n_horiz
                    h_sizes[h] = max(h_sizes[h], im.size[0])
                    v_sizes[v] = max(v_sizes[v], im.size[1])
                h_sizes, v_sizes = np.cumsum(
                    [0] + h_sizes), np.cumsum([0] + v_sizes)
                im_grid = Image.new(
                    'RGB', (h_sizes[-1], v_sizes[-1]), color='white')
                for i, im in enumerate(images):
                    im_grid.paste(
                        im, (h_sizes[i % n_horiz], v_sizes[i // n_horiz]))
                return im_grid
            try:
                return get_grid(min)
            except IndexError:
                return get_grid(max)
        pt = self.folder
        pt = pt / 'state_pdfs'
        name = str(file_name)
        if self.only_pm:
            name = 'all_w_1__' + name
        if not name.endswith('.pdf'):
            name += '.pdf'

        pt.mkdir(parents=True, exist_ok=True)
        filepath = pt / "".join(name)

        whole_image = pil_grid(images, row_len)
        whole_image.save(str(filepath))

        if show:
            plt.figure()
            plt.imshow(whole_image.convert('RGB'))
            plt.axis('off')

    def info_statex(self, idx=0, infos=[], filter_zeros=False, figsize=(14, 8)):

        ### plot setup ###

        fig = plt.figure(str(idx), figsize=figsize)
        graph_ax = fig.add_subplot(1, 2, 1)
        state_ax = fig.add_subplot(2, 2, 2)
        loss_ax = fig.add_subplot(2, 2, 4)

        graph, state = self.turn_dic_in_graph_state(self.files[idx]['graph'])
        st_dic = self.files[idx]
        ### Graph ###

        gp.graphPlot(graph, ax_fig=(fig, graph_ax), show=False)
        try:
            self.summary['target_state'] = 1
        except:
            pass

        ### state + infos ###

        st_string = state.state_string(filter_zeros=filter_zeros)
        print(st_string)
        props = dict(boxstyle='round', facecolor='wheat', alpha=0.5)
        fs = max(19 - st_string.count('\n'), 10)

        text = state_ax.annotate(st_string,
                                 xy=(0.5, 1.05), xycoords=("data", 'axes fraction'),
                                 xytext=(0, -20), textcoords='offset points',
                                 ha="center", va="top", fontsize=fs,
                                 bbox=props)

        props = dict(boxstyle='round', facecolor='grey', alpha=0.2)
        state_ax.annotate(state.info_string(*infos, filter_zeros=filter_zeros),
                          xy=(0.5, 0.), xycoords=text,
                          xytext=(0, -20), textcoords='offset points',
                          ha="center", va="top", fontsize=int(0.8*fs),
                          bbox=props)

        state_ax.axis('off')

        ### loss ###
        loss_history = st_dic['history']
        min_edge = len(st_dic['graph'])
        edges = [ii for ii in range(
            min_edge, min_edge+len(loss_history))]
        edges.reverse()
        for ii in range(len(loss_history[0])):
            loss = [ll[ii] for ll in loss_history]
            loss_ax.plot(
                edges, loss, label=self.loss_func_names[ii], alpha=0.4, lw=3)
            loss_ax.plot(edges, loss, alpha=0.6, lw=0.7, color="black")

        fs_loss_plot = int(fs*0.7)  # make fontsize smaller for loss_plot
        loss_ax.set_ylabel("Loss functions", fontsize=fs_loss_plot)
        loss_ax.set_xlabel("Amount of edges left", fontsize=fs_loss_plot)
        loss_ax.tick_params(axis='x', labelsize=fs_loss_plot)
        loss_ax.tick_params(axis='y', labelsize=fs_loss_plot)
        loss_ax.legend(fontsize=fs_loss_plot)
        loss_ax.invert_xaxis()
        loss_ax.grid()
        plt.tight_layout()
        return state


def input_with_check_ints(question: str, restriciton=[]):
    counter = 0
    while counter < 3:
        try:
            user_input = input(question)
            if '-' not in user_input:
                user_input = [int(user_input)]
            else:
                user_input = [int(xx) for xx in user_input.split('-')]

            if len(restriciton) != 0:

                if all(restriciton[0] <= xx and xx <= restriciton[1] for xx in user_input):
                    return user_input
                else:
                    raise ValueError()

            else:
                return user_input
        except:
            counter += 1
        print(f'inproper input!: try again ({3-counter} trys left)')

    import sys
    sys.exit()


def check_which_summary_file(path):
    folder_list = [xx for xx in path.iterdir() if xx.is_dir()]
    # when only one folder exists we take this folder oc
    if len(folder_list) == 1:
        return folder_list[0]
    for idx, ff in enumerate(folder_list):
        aa = convert_file_path_to_dic(ff / 'summary.json')
        print('\n' + ff.name + f'  ({idx=})' + '\n')
        for key, val in aa.items():
            print(f'  - {key} : {val}')
            last_key = key
        print(len(f'  - {last_key} : {aa[last_key]}')*'-')

    folder_idx = input_with_check_ints(f'Choose one folder index ( 0 - {idx} ):',
                                       restriciton=[0, idx]
                                       )[0]
    return folder_list[folder_idx]


def walk(path):
    for p in Path(path).iterdir():
        if p.is_dir():
            yield p
            yield from walk(p)
            continue


def convert_input_path(path: str):

    # first check if no path given then we have to choose system we want to analyze
    current_dir = os.getcwd()
    if not str(current_dir).endswith('output'):
        output_dir = Path(current_dir) / 'output'
    else:
        output_dir = Path(current_dir)

    if path is None or len(path) == 0:
        folder_list = list(output_dir.iterdir())
        for idx, folder in enumerate(folder_list):

            print(f'( {idx} ) : {folder}'.replace(str(output_dir), ''))
        idx_folder = input_with_check_ints(f'Choose one folder to analyse ( index: 0 - {idx} )\n',
                                           restriciton=[0, idx])[0]
        path = Path(output_dir) / str(folder_list[idx_folder])

    path = Path(path)
    # first check if given path is an absolut path and if exists then return
    for folder in walk(output_dir):
        if str(path) in str(folder):
            path = folder
            break

    if path.is_absolute() and any(pp.suffix == '.json' for pp in path.iterdir()):
        if path.exists():
            return path
        else:
            raise ValueError('Given absolute Path does not exist')

    # second one check if name corresponds to one of the folder in output

    else:
        return check_which_summary_file(path)


def get_analyse(which_directory, all_weights_plus_minus_one=False,
                which_infos=['norm'], create_perfect_machting_pdf=False):

    path = convert_input_path(which_directory)
    a = analyser(path, only_pm=all_weights_plus_minus_one)
    #a.plot_losses()
    indexs = input_with_check_ints(
        f'which state? (int from 0 - {len(a.files)-1} or idx_1 - idx_2): ',
        [0, len(a.files)-1])
    if len(indexs) == 1:
        indexs.append(indexs[0])
    for idx in range(indexs[0],indexs[1]+1):
        a.info_statex(idx, infos=which_infos)
        if create_perfect_machting_pdf:
            a.all_perfect_matchings_to_pdf(idx)
    plt.show()
    input('press arbitary key to exit')
    return 0



def save_graph(file_loc, pm=False, fontsize = 12):
    
    
    def get_figsize(columnwidth):
        """Parameters:
          - wf [float]:  width fraction in columnwidth units
          - hf [float]:  height fraction in columnwidth units.
                         Set by default to golden ratio.
          - columnwidth [float]: width of the column in latex. Get this from LaTeX 
                                 using \showthe\columnwidth
        Returns:  [fig_width,fig_height]: that should be given to matplotlib
        """
        fig_width_pt = columnwidth  # Get this from LaTeX using \showthe\columnwidth
        inches_per_pt = 1.0/72.27               # Convert pt to inch
        golden_mean = (np.sqrt(5)-1.0)/2.0         # Aesthetic ratio
        fig_width = fig_width_pt*inches_per_pt  # width in inches
        fig_height = fig_width*golden_mean      # height in inches
        return [fig_width, fig_height]
    
    def turn_dic_in_graph(graph_dict: dict,only_pm= False, thresholds_amplitudes=np.inf):
        """
        returns a GRaph object and a state_analyzer object for given state

        Parameters
        ----------
        graph_dict : dict
            dict representing a graph.
        thresholds_amplitudes : TYPE, optional
            DESCRIPTION. The default is np.inf.

        Returns
        -------
        TYPE
            DESCRIPTION.

        """
        imaginary = False
        if only_pm:
            for ket, ampl in graph_dict.items():
                if isinstance(ampl,list):
                    imaginary = 'polar'
                    graph_dict[ket] = [1,0] if ampl[0] > 0 else [-1,0]
                else:
                    graph_dict[ket] = 1 if ampl > 0 else -1
        for ket, ampl in graph_dict.items():
            if isinstance(graph_dict[ket],list):
                imaginary = 'polar'
        graph = Graph(graph_dict,imaginary=imaginary)
        if imaginary is not False:
            graph.toCartesian()
        graph.getState()
        return graph
    
    params = {'backend': 'ps',
          'axes.labelsize': 10,
          'legend.fontsize': 10,
          'xtick.labelsize': 8,
          'ytick.labelsize': 8,
          'text.usetex': True,}
    
    plt.rcParams.update(params)

    for folder in [pp for pp in os.listdir(file_loc) if '.' not in pp ]:

        cur_file_loc = Path(file_loc) / folder
        plots = 0
        for xx in os.listdir(cur_file_loc):
            if xx.endswith('.json'):
                plots += 1
        figgy, ax = plt.subplots(1,plots,figsize=get_figsize(418.2555))
    
        for idx,ii in enumerate([xx for xx in os.listdir(cur_file_loc) if xx.endswith('.json')]):

            graph = turn_dic_in_graph(convert_file_path_to_dic(Path(cur_file_loc ) / ii)['graph'],pm)
            try:
                fig = gp.graphPlot(graph, show=False,ax_fig=(figgy, ax[idx] ), fontsize=fontsize,
                               max_thickness= 5, zorder= 20,markersize=10)
            except TypeError:
                fig = gp.graphPlot(graph, show=False,ax_fig=(figgy, ax ), fontsize=fontsize,
                               max_thickness= 5, zorder= 7,markersize=12)
        plt.tight_layout()
        save_path = Path(file_loc ) / (folder + '.pdf')
        fig.savefig(str(save_path), format='pdf', bbox_inches='tight')
            
if __name__ == '__main__':
    save_graph(r'C:/Users/janpe/Downloads/test_th/paper',pm=True)
  



import matplotlib.pyplot as plt
import numpy as np
import matplotlib.collections as collections
import pytheus.theseus as th
import pytheus.analyzer as anal
import matplotlib.patheffects as pe
from pytheus.fancy_classes import Graph
import json
import os
import pytheus.leiwand


def drawEdge(edge, verts, ind, mult, ax, scale_max=None, max_thickness=10,
             show_val=False, fs=15, markersize=25):
    colors = ['dodgerblue', 'firebrick', 'limegreen', 'darkorange', 'purple', 'yellow', 'cyan']
    col1 = colors[int(edge[2])]
    col2 = colors[int(edge[3])]

    vert1 = np.array(verts[int(edge[0])])
    vert2 = np.array(verts[int(edge[1])])
    if not np.array_equal(vert1, vert2):
        diff = vert1 - vert2
        rect = [diff[1], -diff[0]]
        rect /= np.linalg.norm(rect)
        hp = (vert1 + vert2) / 2 + (2 * ind - mult + 1) * 0.05 * rect
    else:
        hp = vert1 * 1.2

    if scale_max is None:
        lw = max_thickness

    else:
        lw = np.max([abs(max_thickness * edge[4]) / scale_max, 0.5])

    try:
        transparency = 0.2 + abs(edge[4]) * 0.8
        transparency = min(transparency, 1)
    except IndexError:
        transparency = 1
    except TypeError:
        transparency = 1

    ax.plot([vert1[0], hp[0]], [vert1[1], hp[1]], color=col1, linewidth=lw, alpha=transparency)
    ax.plot([hp[0], vert2[0]], [hp[1], vert2[1]], col2, linewidth=lw, alpha=transparency)

    if show_val:

        if transparency > 0.5 and col1 == "blue":
            font_col = 'white'
        else:
            font_col = 'black'
        latex_weight = '${}$'.format(anal.num_in_str(edge[4]))
        if latex_weight == '$$':
            latex_weight = str(edge[4])
        ax.text(np.mean([0.9 * vert1[0], hp[0]]), np.mean([0.9 * vert1[1], hp[1]]),
                latex_weight,
                bbox={'facecolor': col1, 'alpha': transparency, 'edgecolor': col2, 'pad': 1}, c=font_col,
                ha='center', va='center', rotation=0, fontweight='heavy', fontsize=fs)
    try:
        if edge[4] < 0:
            ax.plot(hp[0], hp[1], marker="d", markersize=markersize, markeredgewidth="3", markeredgecolor="black",
                    color="white")
    except:
        pass


def graphPlot(graph, scaled_weights=False, show=True, max_thickness=10,
              weight_product=False, ax_fig=(), add_title='',
              show_value_for_each_edge=False, fontsize=30, zorder=11,
              markersize=25, number_nodes=True, filename=''):
    edge_dict = th.edgeBleach(graph.edges)

    num_vertices = len(np.unique(np.array(graph.edges)[:, :2]))

    angles = np.linspace(0, 2 * np.pi * (num_vertices - 1) / num_vertices, num_vertices)

    rad = 0.9
    vertcoords = []
    for angle in angles:
        x = rad * np.cos(angle)
        y = rad * np.sin(angle)
        vertcoords.append(tuple([x, y]))

    vertnums = list(range(num_vertices))
    verts = dict(zip(vertnums, vertcoords))

    if scaled_weights:
        try:
            scale_max = np.max(np.abs(np.array(graph.edges)[:, 4]))
        except:
            scale_max = None
    else:
        scale_max = None

    if len(ax_fig) == 0:
        fig, ax = plt.subplots(figsize=(10, 10))
    else:
        fig, ax = ax_fig

    for uc_edge in edge_dict.keys():
        mult = len(edge_dict[uc_edge])
        for ii, coloring in enumerate(edge_dict[uc_edge]):
            drawEdge(uc_edge + coloring + tuple([graph[tuple(uc_edge + coloring)]]), verts, ii, mult, ax,
                     scale_max=scale_max, max_thickness=max_thickness,
                     show_val=show_value_for_each_edge, fs=0.8 * fontsize, markersize=markersize)

    circ = []
    if number_nodes:
        node_labels = verts.keys()
    else:
        node_labels = list(map(chr, range(97, 123)))
    for vert, coords in zip(node_labels, verts.values()):
        circ.append(plt.Circle(coords, 0.1, alpha=0.5))
        ax.text(coords[0], coords[1], str(vert), zorder=zorder,
                ha='center', va='center', size=fontsize)

    circ = collections.PatchCollection(circ, zorder=zorder - 1)
    circ.set(facecolor='lightgrey', edgecolor='dimgray', linewidth=3)
    ax.add_collection(circ)

    ax.set_xlim([-1.1, 1.1])
    ax.set_ylim([-1.1, 1.1])
    ax.axis('off')

    if weight_product:
        total_weight = np.product(graph.weights)

        wp = '${}$'.format(anal.num_in_str(total_weight))
        if wp == '$$':
            wp = str(total_weight)
        ax.set_title(wp + str(add_title), fontsize=fontsize)

    if add_title != '' and weight_product is False:
        ax.set_title(str(add_title), fontsize=fontsize)

    if show:
        plt.show()
        plt.pause(0.01)
    else:
        pass
    if filename:
        fig.savefig(filename + ".pdf")

    return fig


def leiwandPlot(graph, name='graph'):
    data = []
    edge_dict = th.edgeBleach(graph.edges)
    for uc_edge in edge_dict.keys():
        mult = len(edge_dict[uc_edge])
        loop = (uc_edge[0] == uc_edge[1])
        for ii, coloring in enumerate(edge_dict[uc_edge]):
            edge = tuple(uc_edge + coloring)
            weight = graph[edge]
            if loop:
                loose = 10 + 5 * ii
                data.append([weight, str(edge[0]), edge[2], str(edge[1]), edge[3], loose])
            else:
                bend = -22.5 + (ii + 0.5) * 45 / mult
                data.append([weight, str(edge[0]), edge[2], str(edge[1]), edge[3], bend])
    pytheus.leiwand.leiwand(data, name)


def leiwandPlotBulk(graph, cnfg, root, name = 'graph'):
    # if graph is imaginary, just take absolute value as weight for now
    if graph.imaginary:
        graph.absolute()
    data = []
    edge_dict = th.edgeBleach(graph.edges)
    for uc_edge in edge_dict.keys():
        mult = len(edge_dict[uc_edge])
        loop = (uc_edge[0] == uc_edge[1])
        for ii, coloring in enumerate(edge_dict[uc_edge]):
            edge = tuple(uc_edge + coloring)
            weight = graph[edge]
            if loop:
                loose = 10 + 5 * ii
                data.append([weight, str(edge[0]), edge[2], str(edge[1]), edge[3], loose])
            else:
                bend = -22.5 + (ii + 0.5) * 45 / mult
                data.append([weight, str(edge[0]), edge[2], str(edge[1]), edge[3], bend])
    pytheus.leiwand.leiwandBulk(data, cnfg, root=root, name=name)


def plotFromFile(filename, number_nodes=True, outfile=""):
    if not os.path.exists(filename) or os.path.isdir(filename):
        raise IOError(f'File does not exist: {filename}')
    with open(filename) as input_file:
        sol_dict = json.load(input_file)
    graph = Graph(sol_dict['graph'])
    graphPlot(graph, scaled_weights=True, number_nodes=number_nodes, filename=outfile)



import os
import tempfile
import unittest
from pathlib import Path

from click.testing import CliRunner

from pytheus.cli import run


class FunctionalTests(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        print('Running functional tests. This may take a while.')

    def setUp(self):
        self.directory = tempfile.TemporaryDirectory()
        os.chdir(self.directory.name)

    def tearDown(self):
        self.directory.cleanup()

    def test_input_with_json_ending_from_example_dir(self):
        runner = CliRunner()
        result = runner.invoke(run, ['--example', 'ghz_346.json'])
        assert result.exit_code == 0
        assert "{'|000000>': True, '|111000>': True, '|222000>': True, '|333000>': True}" in result.output
        assert os.path.exists('output/ghz_346/ghz_346/best.json')
        assert os.path.exists('output/ghz_346/ghz_346/summary.json')

    def test_input_without_json_ending_from_example_dir(self):
        runner = CliRunner()
        result = runner.invoke(run, ['--example', 'ghz_346'])
        assert result.exit_code == 0
        assert "{'|000000>': True, '|111000>': True, '|222000>': True, '|333000>': True}" in result.output
        assert os.path.exists('output/ghz_346/ghz_346/best.json')
        assert os.path.exists('output/ghz_346/ghz_346/summary.json')

    def test_input_with_json_ending_from_custom_dir(self):
        input_file = Path(__file__).parent / 'fixtures' / 'ghz_346.json'
        runner = CliRunner()
        result = runner.invoke(run, [str(input_file)])
        assert result.exit_code == 0
        assert "{'|000000>': True, '|111000>': True, '|222000>': True, '|333000>': True}" in result.output
        assert os.path.exists('output/ghz_346/ghz_346/best.json')
        assert os.path.exists('output/ghz_346/ghz_346/summary.json')

    def test_input_without_json_ending_from_custom_dir(self):
        input_file = Path(__file__).parent / 'fixtures' / 'ghz_346'
        runner = CliRunner()
        result = runner.invoke(run, [str(input_file)])
        assert result.exit_code == 0
        assert "{'|000000>': True, '|111000>': True, '|222000>': True, '|333000>': True}" in result.output
        assert os.path.exists('output/ghz_346/ghz_346/best.json')
        assert os.path.exists('output/ghz_346/ghz_346/summary.json')

    def test_non_existing_input_stops_with_error_message(self):
        runner = CliRunner()
        result = runner.invoke(run, ['i_dont_exist.json'])
        assert result.exit_code == 1
        assert 'ERROR' in result.output

    def test_bell_state(self):
        input_file = Path(__file__).parent / 'fixtures' / 'bell'
        runner = CliRunner()
        result = runner.invoke(run, [str(input_file)])
        assert 'finished with graph with 2 edges' in result.output

    def test_lossfunc_ent(self):
        runner = CliRunner()
        result = runner.invoke(run, ['--example', 'conc_4-3'])
        assert result.exit_code == 0
        assert os.path.exists('output/conc_4-3/try/best.json')
        assert os.path.exists('output/conc_4-3/try/summary.json')

    def test_input_with_json_ending_from_example_director_removeConnections(self):
        runner = CliRunner()
        result = runner.invoke(run, ['--example', 'cnot_22.json'])
        assert result.exit_code == 0

    def test_input_with_json_ending_from_example_director_startinggraph(self):
        runner = CliRunner()
        result = runner.invoke(run, ['--example', 'spin3plus_3.json'])
        assert result.exit_code == 0

    def test_input_with_json_ending_from_example1_director(self):
        runner = CliRunner()
        result = runner.invoke(run, ['--example', 'werner.json'])
        assert result.exit_code == 0


import itertools
import time
import copy
import random
from random import shuffle
import numpy as np
from scipy import optimize
import pytheus.fancy_classes as fc
import pytheus.theseus as th


def flatten(l):
    return [item for sublist in l for item in sublist]


def is_connected(lst):
    # print('lst: ',lst)
    in_list = [lst[0][0], lst[0][1]]

    cnum_vertices = len(set(flatten([[vv[0], vv[1]] for vv in lst])))

    curr_len = len(in_list) - 1
    # print('len(in_list)<cnum_vertices: ',len(in_list)<cnum_vertices)
    # print('len(in_list)>curr_len: ',len(in_list)>curr_len)

    while len(in_list) < cnum_vertices and len(in_list) > curr_len:
        curr_len = len(in_list)
        for ee in lst[1:]:
            if (ee[0] in in_list) and (not ee[1] in in_list):
                in_list.append(ee[1])
            if (ee[1] in in_list) and (not ee[0] in in_list):
                in_list.append(ee[0])

    # print(in_list)
    # print(cnum_vertices)
    return len(in_list) == cnum_vertices


def make_list_unique(lst):
    unique_lst = [list(x) for x in set(tuple(x) for x in lst)]
    return unique_lst


def is_substructure(lst1, lst2):
    if len(lst1) < len(lst2):
        return False

    rr = [ll in lst1 for ll in lst2]
    if not all(rr):
        return False

    unique_lst2 = make_list_unique(lst2)
    if len(lst2) > len(unique_lst2):
        return False

    return True


def compute_all_possibilies(full_graph, all_curr_subsequence, num_vertices, num_cols):
    # print('all_curr_subsequence: ', all_curr_subsequence)
    # time.sleep(0.5)
    all_permutations = list(itertools.permutations(list(range(num_vertices))))
    all_permutations_cols = list(itertools.permutations(list(range(num_cols))))
    curr_graph = all_curr_subsequence[-1]

    # print('len(all_permutations): ', len(all_permutations))

    all_curr_subsequence_ext = all_curr_subsequence + [[ee] for ee in full_graph]
    # print('len(all_curr_subgraphs_ext): ', len(all_curr_subsequence_ext))
    all_possibilities = []
    for curr_substr in all_curr_subsequence_ext:
        if len(curr_substr) == 1:
            new_graph = curr_graph + curr_substr
            if is_substructure(full_graph, new_graph):
                if not new_graph in all_possibilities:
                    all_possibilities.append(new_graph)

        if len(curr_substr) > 1:
            if is_connected(curr_substr):
                # print('curr_substr: ',curr_substr)
                for curr_perm in all_permutations:
                    # print('all_permutations: ', all_permutations)

                    # print('curr_substr: ',curr_substr)
                    for curr_col_perm in all_permutations_cols:
                        curr_substr_perm = []
                        for cedge in curr_substr:
                            nedge = [curr_perm[cedge[0]], curr_perm[cedge[1]], curr_col_perm[cedge[2]],
                                     curr_col_perm[cedge[3]]]
                            # print('nedge: ', nedge)
                            if nedge[0] > nedge[1]:
                                nedge = [nedge[1], nedge[0], nedge[2], nedge[3]]
                            curr_substr_perm.append(nedge)
                            # time.sleep(0.1)

                        new_graph = curr_graph + curr_substr_perm
                        # print('  new_graph: ',new_graph, '(curr_perm: ', curr_perm,')')

                        # time.sleep(0.1)
                        if is_substructure(full_graph, new_graph):
                            # print('is_substructure')
                            if not new_graph in all_possibilities:
                                # print('not new_graph in all_possibilities')
                                all_possibilities.append(new_graph)

    return all_possibilities


def compute_assembly_index(full_graph, all_curr_subsequence, assembly_index_col, num_vertices, num_cols):
    # print(' - - - - -')
    # print('in compute_assembly_index')
    # print('  full_graph: ',full_graph)
    # for csg in all_curr_subsequence:
    #    print('    csg: ', csg)
    # time.sleep(0.25)

    if len(full_graph) == 0:
        return 0, []
    if len(full_graph) == 1:
        return 1, [[full_graph[0]]]
    if len(full_graph) == 2:
        return 2, [[full_graph[0]], full_graph]

    global min_assembly_idx, min_ai_structure, assembly_index_collection
    num_vertices = max([max(ll[0:2]) for ll in full_graph]) + 1
    num_cols = max([max(ll[2:4]) for ll in full_graph]) + 1

    all_possibilies = compute_all_possibilies(full_graph, all_curr_subsequence, num_vertices, num_cols)
    # print('len(all_possibilies): ', len(all_possibilies))
    for new_graph in all_possibilies:
        new_curr_subgraphs = copy.deepcopy(all_curr_subsequence)
        new_curr_subgraphs.append(new_graph)
        # print('new_graph: ', new_graph)
        # time.sleep(1)

        if len(new_curr_subgraphs) < min_assembly_idx:
            if is_substructure(full_graph, new_graph) and is_substructure(new_graph, full_graph):
                # print(' DONE !!! Assembly Index: ', len(new_curr_subgraphs))
                assembly_index_col.append(len(new_curr_subgraphs))
                if len(new_curr_subgraphs) < min_assembly_idx:
                    min_assembly_idx = len(new_curr_subgraphs)
                    min_ai_structure = new_curr_subgraphs
                    # print('new best value: ', min_assembly_idx)
            else:
                compute_assembly_index(full_graph, new_curr_subgraphs, assembly_index_collection, num_vertices,
                                       num_cols)

    return min_assembly_idx, min_ai_structure


def assembly_index_unweighted(gg, num_vertices, num_cols):
    global min_assembly_idx, min_ai_structure, assembly_index_collection
    min_assembly_idx = 666
    min_ai_structure = []
    assembly_index_collection = []

    if len(gg) == 0:
        # print('assembly_index_collection: ', 0)
        return 0

    init_structure = [gg[0]]
    # print(gg)

    min_assembly_idx, min_ai_structure = compute_assembly_index(gg, [init_structure], assembly_index_collection,
                                                                num_vertices, num_cols)
    # print('assembly_index_collection: ', min_assembly_idx)
    # for ii in min_ai_structure:
    # print(ii)

    return min_assembly_idx


def sample_subgraph(graph, size_of_graph):
    all_edges = graph.edges
    all_weights = graph.weights

    curr_edges = []
    while len(curr_edges) < size_of_graph:
        ridx = random.randint(0, len(all_edges) - 1)
        if random.random() < abs(all_weights[ridx]):
            if not all_edges[ridx] in curr_edges:
                curr_edges.append(all_edges[ridx])

    curr_graph = fc.Graph(curr_edges)
    for edge in curr_edges:
        curr_graph[edge] = graph[edge]

    return curr_graph


def assembly_index(graph, cnfg):
    print("computing assembly index")
    num_vertices = cnfg["num_vertices"]
    num_cols = cnfg["num_cols"]
    size_of_graph = cnfg["size_of_graph"]

    all_sampled_assembly_indices = []

    for ii in range(cnfg["sample_size"]):
        sampled_graph = sample_subgraph(graph, size_of_graph)
        sampled_graph = sampled_graph.edges
        sampled_graph = [list(edge) for edge in sampled_graph]
        sampled_graph.sort()
        min_assembly_idx = assembly_index_unweighted(sampled_graph, num_vertices, num_cols)
        all_sampled_assembly_indices.append(min_assembly_idx)

    weighted_assembly_index = sum(all_sampled_assembly_indices) / len(all_sampled_assembly_indices)
    return weighted_assembly_index


def sample_top(graph, cnfg, ii):
    sorted_inds = list(np.argsort(graph.weights))
    sorted_inds.reverse()
    sorted_edges = [graph.edges[ind] for ind in sorted_inds]
    # weight of the edge that gets promoted
    weight = graph[sorted_edges[ii + cnfg["size_of_graph"]]]
    sampled_edges = sorted_edges[:cnfg["size_of_graph"] - 1] + [sorted_edges[ii + cnfg["size_of_graph"]]]
    sampled_edges = [list(edge) for edge in sampled_edges]
    sampled_edges.sort()
    return sampled_edges, weight


def sample_bottom(graph, cnfg, ii):
    sorted_inds = list(np.argsort(graph.weights))
    sorted_inds.reverse()
    sorted_edges = [graph.edges[ind] for ind in sorted_inds]
    # weight of the edge that gets demoted
    weight = graph[sorted_edges[ii]]
    sampled_edges = sorted_edges[:ii] + sorted_edges[ii + 1:cnfg["size_of_graph"]] + [sorted_edges[
                                                                                          cnfg["size_of_graph"] + ii]]
    sampled_edges = [list(edge) for edge in sampled_edges]
    sampled_edges.sort()
    return sampled_edges, weight


def top_n_assembly(graph, cnfg):
    print("computing top_n_assembly loss")
    num_vertices = cnfg["num_vertices"]
    num_cols = cnfg["num_cols"]
    size_of_graph = cnfg["size_of_graph"]

    for e in graph.edges:
        graph[e] = np.abs(graph[e])

    sorted_inds = list(np.argsort(graph.weights))
    sorted_inds.reverse()
    sorted_edges = [graph.edges[ind] for ind in sorted_inds]
    top_edges = sorted_edges[:cnfg["size_of_graph"]]
    top_edges = [list(edge) for edge in top_edges]
    top_edges.sort()
    curr_assembly = assembly_index_unweighted(top_edges, num_vertices, num_cols)

    lossfunc = 0

    for ii in range(len(graph) - cnfg["size_of_graph"]):
        # check the assembly index if smallest of top edges was switched out for each of the bottom edges
        # get weight of bottom edge to be promoted
        sampled_edges, weight = sample_top(graph, cnfg, ii)
        sample_assembly = assembly_index_unweighted(sampled_edges, num_vertices, num_cols)
        lossfunc += (sample_assembly - curr_assembly) * weight

    for ii in range(cnfg["size_of_graph"]):
        # check the assembly index if biggest of bottom edges was switched out for each of the top edges
        # get weight of top edge to be demoted
        sampled_edges, weight = sample_bottom(graph, cnfg, ii)
        sample_assembly = assembly_index_unweighted(sampled_edges, num_vertices, num_cols)
        lossfunc += (curr_assembly - sample_assembly) * weight

    return lossfunc


if __name__ == "__main__":
    cnfg = {
        "num_vertices": 4,
        "num_cols": 2,
        "size_of_graph": 8}
    gg = fc.Graph(th.buildAllEdges([2, 2, 2, 2]))
    for e in gg.edges:
        gg[e] = random.random()
    ai = top_n_assembly(gg, cnfg)
    print(ai)



import abc
import json
import sys
import traceback
import unittest
from filecmp import cmp
from typing import List

import numpy as np
from numpy import array

from pytheus import main
from pytheus.help_functions import readableState
from pytheus.main import read_config, get_dimensions_and_target_state, build_starting_graph, setup_for_ent, \
    setup_for_target, setup_for_fockbasis, optimize_graph, run_main


class TestMainModule(unittest.TestCase):

    def test_read_config_from_example_dir_with_json_ending(self):
        config, filename = read_config(is_example=True, filename='ghz_346.json')

        self.assertEqual(
            config['target_state'], ["000", "111", "222", "333"]
        )

    def test_read_config_from_example_dir_without_json_ending(self):
        config, filename = read_config(is_example=True, filename='ghz_346')

        self.assertEqual(
            config['target_state'], ["000", "111", "222", "333"]
        )
        self.assertEqual('ghz_346.json', filename.name)

    def test_build_starting_graph(self):
        cnfg, filename = read_config(is_example=True, filename='ghz_346')
        dimension_key = [4, 4, 4, 1, 1, 1]
        expected_outcome = {(0, 1, 0, 0): True, (0, 1, 0, 1): True, (0, 1, 0, 2): True, (0, 1, 0, 3): True,
                            (0, 1, 1, 0): True, (0, 1, 1, 1): True, (0, 1, 1, 2): True, (0, 1, 1, 3): True,
                            (0, 1, 2, 0): True, (0, 1, 2, 1): True, (0, 1, 2, 2): True, (0, 1, 2, 3): True,
                            (0, 1, 3, 0): True, (0, 1, 3, 1): True, (0, 1, 3, 2): True, (0, 1, 3, 3): True,
                            (0, 2, 0, 0): True, (0, 2, 0, 1): True, (0, 2, 0, 2): True, (0, 2, 0, 3): True,
                            (0, 2, 1, 0): True, (0, 2, 1, 1): True, (0, 2, 1, 2): True, (0, 2, 1, 3): True,
                            (0, 2, 2, 0): True, (0, 2, 2, 1): True, (0, 2, 2, 2): True, (0, 2, 2, 3): True,
                            (0, 2, 3, 0): True, (0, 2, 3, 1): True, (0, 2, 3, 2): True, (0, 2, 3, 3): True,
                            (0, 3, 0, 0): True, (0, 3, 1, 0): True, (0, 3, 2, 0): True, (0, 3, 3, 0): True,
                            (0, 4, 0, 0): True, (0, 4, 1, 0): True, (0, 4, 2, 0): True, (0, 4, 3, 0): True,
                            (0, 5, 0, 0): True, (0, 5, 1, 0): True, (0, 5, 2, 0): True, (0, 5, 3, 0): True,
                            (1, 2, 0, 0): True, (1, 2, 0, 1): True, (1, 2, 0, 2): True, (1, 2, 0, 3): True,
                            (1, 2, 1, 0): True, (1, 2, 1, 1): True, (1, 2, 1, 2): True, (1, 2, 1, 3): True,
                            (1, 2, 2, 0): True, (1, 2, 2, 1): True, (1, 2, 2, 2): True, (1, 2, 2, 3): True,
                            (1, 2, 3, 0): True, (1, 2, 3, 1): True, (1, 2, 3, 2): True, (1, 2, 3, 3): True,
                            (1, 3, 0, 0): True, (1, 3, 1, 0): True, (1, 3, 2, 0): True, (1, 3, 3, 0): True,
                            (1, 4, 0, 0): True, (1, 4, 1, 0): True, (1, 4, 2, 0): True, (1, 4, 3, 0): True,
                            (1, 5, 0, 0): True, (1, 5, 1, 0): True, (1, 5, 2, 0): True, (1, 5, 3, 0): True,
                            (2, 3, 0, 0): True, (2, 3, 1, 0): True, (2, 3, 2, 0): True, (2, 3, 3, 0): True,
                            (2, 4, 0, 0): True, (2, 4, 1, 0): True, (2, 4, 2, 0): True, (2, 4, 3, 0): True,
                            (2, 5, 0, 0): True, (2, 5, 1, 0): True, (2, 5, 2, 0): True, (2, 5, 3, 0): True,
                            (3, 4, 0, 0): True, (3, 5, 0, 0): True, (4, 5, 0, 0): True}
        actual = build_starting_graph(cnfg, dimension_key)
        self.assertEqual(87, len(actual))
        self.assertEqual(dimension_key, actual.dimensions)
        self.assertEqual(expected_outcome, actual.graph)
        self.assertEqual(list(expected_outcome.values()), actual.weights)

    def test_get_dimensions_and_target_state(self):
        cnfg, filename = read_config(is_example=True, filename='ghz_346')
        exp = ([4, 4, 4, 1, 1, 1], None, {((0, 0), (1, 0), (2, 0), (3, 0), (4, 0), (5, 0)): True,
                                          ((0, 1), (1, 1), (2, 1), (3, 0), (4, 0), (5, 0)): True,
                                          ((0, 2), (1, 2), (2, 2), (3, 0), (4, 0), (5, 0)): True,
                                          ((0, 3), (1, 3), (2, 3), (3, 0), (4, 0), (5, 0)): True})
        actual = get_dimensions_and_target_state(cnfg)
        self.assertEqual([4, 4, 4, 1, 1, 1], actual[0])
        self.assertIsNone(actual[1])
        self.assertEqual(list(exp[2].values()), actual[2].amplitudes)
        self.assertEqual(list(exp[2].keys()), actual[2].kets)

    def test_setup_for_ent(self):
        cnfg, filename = read_config(is_example=True, filename='conc_4-3')
        exp = ([2, 2, 2, 2], {'dimensions': [2, 2, 2, 2], 'num_ancillas': 0, 'num_particles': 4,
                              'all_states': [((0, 0), (1, 0), (2, 0), (3, 0)), ((0, 0), (1, 0), (2, 0), (3, 1)),
                                             ((0, 0), (1, 0), (2, 1), (3, 0)), ((0, 0), (1, 0), (2, 1), (3, 1)),
                                             ((0, 0), (1, 1), (2, 0), (3, 0)), ((0, 0), (1, 1), (2, 0), (3, 1)),
                                             ((0, 0), (1, 1), (2, 1), (3, 0)), ((0, 0), (1, 1), (2, 1), (3, 1)),
                                             ((0, 1), (1, 0), (2, 0), (3, 0)), ((0, 1), (1, 0), (2, 0), (3, 1)),
                                             ((0, 1), (1, 0), (2, 1), (3, 0)), ((0, 1), (1, 0), (2, 1), (3, 1)),
                                             ((0, 1), (1, 1), (2, 0), (3, 0)), ((0, 1), (1, 1), (2, 0), (3, 1)),
                                             ((0, 1), (1, 1), (2, 1), (3, 0)), ((0, 1), (1, 1), (2, 1), (3, 1))],
                              'dim_total': 16, 'bipar_for_opti': [([0, 1], [2, 3]), ([1, 2], [0, 3]), ([0, 2], [1, 3])],
                              'imaginary': False}, {(0, 1, 0, 0): True, (0, 1, 0, 1): True, (0, 1, 1, 0): True,
                                                    (0, 1, 1, 1): True, (0, 2, 0, 0): True, (0, 2, 0, 1): True,
                                                    (0, 2, 1, 0): True, (0, 2, 1, 1): True, (0, 3, 0, 0): True,
                                                    (0, 3, 0, 1): True, (0, 3, 1, 0): True, (0, 3, 1, 1): True,
                                                    (1, 2, 0, 0): True, (1, 2, 0, 1): True, (1, 2, 1, 0): True,
                                                    (1, 2, 1, 1): True, (1, 3, 0, 0): True, (1, 3, 0, 1): True,
                                                    (1, 3, 1, 0): True, (1, 3, 1, 1): True, (2, 3, 0, 0): True,
                                                    (2, 3, 0, 1): True, (2, 3, 1, 0): True, (2, 3, 1, 1): True})
        actual = setup_for_ent(cnfg)
        self.assertEqual([2, 2, 2, 2], actual[0])
        self.assertEqual(exp[1], actual[1])
        self.assertEqual(list(exp[1].values()), list(actual[1].values()))
        self.assertEqual(exp[1].keys(), actual[1].keys())
        self.assertEqual(list(exp[2].values()), actual[2].weights)
        self.assertEqual(list(exp[2].keys()), actual[2].edges)

    @unittest.skip
    #added some features to the config. is it possible to check if config contains out_config, so we dont get failing
    #tests if we expand the config features more in the future
    def test_setup_for_target(self):
        cnfg, filename = read_config(is_example=True, filename='cnot_22.json')
        read_state = {'|000000>': True, '|010100>': True, '|101100>': True, '|111000>': True}
        kets = [((0, 0), (1, 0), (2, 0), (3, 0), (4, 0), (5, 0)), ((0, 0), (1, 1), (2, 0), (3, 1), (4, 0), (5, 0)),
                ((0, 1), (1, 0), (2, 1), (3, 1), (4, 0), (5, 0)), ((0, 1), (1, 1), (2, 1), (3, 0), (4, 0), (5, 0))]
        out_config = {'description': 'Postselected CNOT between two qubits. Two ancillary particles from SPDC.',
                      'edges_tried': 30, 'foldername': 'cnot_22', 'ftol': 1e-06, 'loss_func': 'cr', 'num_anc': 2,
                      'optimizer': 'L-BFGS-B', 'imaginary': False, 'safe_hist': True, 'samples': 10,
                      'target_state': ['0000', '0101', '1011', '1110'], 'in_nodes': [0, 1], 'out_nodes': [2, 3],
                      'heralding_out': True, 'novac': True, 'thresholds': [0.3, 0.1], 'tries_per_edge': 5,
                      'topopt': True, 'single_emitters': [], 'removed_connections': [[0, 1]], 'unicolor': False,
                      'amplitudes': [], 'number_resolving': False, 'brutal_covers': False, 'bulk_thr': 0,
                      'save_hist': True, 'num_pre': 1, 'dimensions': [2, 2, 2, 2, 1, 1],
                      'verts': array([0, 1, 2, 3, 4, 5]), 'anc_detectors': [4, 5]}
        graph = {(0, 2, 0, 0): True, (0, 2, 0, 1): True, (0, 2, 1, 0): True, (0, 2, 1, 1): True, (0, 3, 0, 0): True,
                 (0, 3, 0, 1): True, (0, 3, 1, 0): True, (0, 3, 1, 1): True, (0, 4, 0, 0): True, (0, 4, 1, 0): True,
                 (0, 5, 0, 0): True, (0, 5, 1, 0): True, (1, 2, 0, 0): True, (1, 2, 0, 1): True, (1, 2, 1, 0): True,
                 (1, 2, 1, 1): True, (1, 3, 0, 0): True, (1, 3, 0, 1): True, (1, 3, 1, 0): True, (1, 3, 1, 1): True,
                 (1, 4, 0, 0): True, (1, 4, 1, 0): True, (1, 5, 0, 0): True, (1, 5, 1, 0): True, (2, 3, 0, 0): True,
                 (2, 3, 0, 1): True, (2, 3, 1, 0): True, (2, 3, 1, 1): True, (2, 4, 0, 0): True, (2, 4, 1, 0): True,
                 (2, 5, 0, 0): True, (2, 5, 1, 0): True, (3, 4, 0, 0): True, (3, 4, 1, 0): True, (3, 5, 0, 0): True,
                 (3, 5, 1, 0): True, (4, 5, 0, 0): True}
        actual = setup_for_target(cnfg)
        self.assertEqual(list(kets), list(actual[0].kets))
        self.assertTrue(all(actual[0].amplitudes))
        # self.assertEqual(read_state, readableState(actual[0]))
        self.assertEqual(list(graph.values()), actual[1].weights)
        self.assertEqual(list(graph.keys()), actual[1].edges)
        self.assertEqual(out_config.keys(), actual[2].keys())
        self.assertSetEqual(set(map(type, out_config.values())), set(map(type, actual[2].values())))
        self.assertEqual(all(out_config.values()), all(actual[2].values()))

    def test_setup_for_fockbasis(self):
        cnfg, filename = read_config(is_example=True, filename='fock_tetrahedron_short.json')
        actual = setup_for_fockbasis(cnfg)
        self.assertEqual([((0, 0), (0, 0), (0, 0), (2, 0)), ((1, 0), (1, 0), (1, 0), (2, 0))], actual[0].kets)
        self.assertEqual([1, 1.4142135623730951], actual[0].amplitudes)
        self.assertEqual([1, 1, 1], actual[1])
        self.assertIsNone(actual[2])
        self.assertEqual([(0, 0, 0, 0), (0, 1, 0, 0), (0, 2, 0, 0), (1, 1, 0, 0), (1, 2, 0, 0), (2, 2, 0, 0)],
                         actual[3].edges)
        self.assertTrue(all(actual[3].weights))

    @unittest.skip
    #this fails because the results will generally vary with every run. we did implement a 'seed' option for the config files.
    #when seed is set the result of the first sample will always be the same
    def test_optimize_graph(self):
        cnfg, filename = read_config(is_example=True, filename='werner.json')
        dimension = [2, 2, 5, 1]
        exp_output = {(0, 1, 1, 1): -0.4115376348306805, (1, 2, 0, 1): 0.43097397912168994,
                      (1, 2, 1, 2): 0.4309739796322021, (0, 2, 1, 3): 0.44939847729263027,
                      (1, 3, 0, 0): -0.8812150910946458, (0, 3, 0, 0): -0.9185771000730444,
                      (1, 2, 1, 0): 0.9592641504673297, (2, 3, 4, 0): 0.962075654675881, (0, 2, 1, 0): 1.0}
        t_state = setup_for_target(cnfg)
        np.random.seed(0)
        actual = optimize_graph(cnfg, dimension, filename, build_starting_graph(cnfg, dimension), None, t_state[0])
        self.assertEqual([5,5,5,1], actual.dimensions)
        self.assertEqual(9, len(actual))
        self.assertEqual(exp_output, actual.graph)



# -*- coding: utf-8 -*-
"""
Created on Tue Jul 12 12:24:53 2022

@author: janpe
"""

import json
import os.path
import sys
from pathlib import Path

import pkg_resources
import logging

log = logging.getLogger(__name__)

import pytheus
import pytheus.help_functions as hf
import pytheus.saver as saver
import pytheus.theseus as th
from pytheus.fancy_classes import Graph, State
from pytheus.optimizer import topological_opti
import itertools
import numpy as np
import random


def run_main(filename, example, run_opt=True, state_cat=True):
    """Run the Theseus algorithm on a given input file.

    Parameters
    ----------
    filename: str
        case name or input file path
    example: bool
        flag indicating whether to run included example case or external file.


    Raises
    ------
    IOError
        if filename is not valid.
    """
    # step 1: read in config file
    cnfg, filename = read_config(example, filename)
    logging.basicConfig(stream=sys.stdout, level=logging.INFO)
    logging.info(filename)
    if 'description' in cnfg.keys():
        logging.info(cnfg['description'])

    sys.setrecursionlimit(1000000000)

    # step 2: build up target and starting graph
    if cnfg['loss_func'] == 'ent':  # optimization for entanglement requires specific setup
        dimensions, sys_dict, start_graph = setup_for_ent(cnfg)
        target_state = None
    elif cnfg['loss_func'] == 'lff':  # optimization of a custom loss function
        edge_list = th.buildAllEdges(cnfg["dimensions"], imaginary=cnfg['imaginary'])
        print(f'start graph has {len(edge_list)} edges.')
        start_graph = Graph(edge_list, imaginary=cnfg['imaginary'])
        dimensions = cnfg["dimensions"]
        target_state = None
        sys_dict = None
    elif cnfg['loss_func'] in ['fockcr', 'fockfid']:
        # ADD SETUP FOR FOCK OPTIMIZATION HERE
        # start_graph, target_state, dimensions = setup_for_fockbasis()
        sys_dict = None
        target_state, dimensions, sys_dict, start_graph = setup_for_fockbasis(cnfg)

    else:  # optimization for target given in config
        # read out target and starting graph from cnfg
        # modifies cnfg to incorporate topological constraints
        target_state, start_graph, cnfg = setup_for_target(cnfg, state_cat=state_cat)
        # target_state is state object
        # start_graph is graph object
        dimensions = cnfg["dimensions"]
        sys_dict = None

    # step 3: start optimization
    if run_opt:
        optimize_graph(cnfg, dimensions, filename, start_graph, sys_dict, target_state)
    else:
        return cnfg


def optimize_graph(cnfg, dimensions, filename, start_graph, sys_dict, target_state):
    '''
    main optimization routine

    Parameters
    ----------
    cnfg
    dimensions
    filename
    start_graph
    sys_dict
    target_state

    Returns
    -------
    result graph
    '''
    # initialize saver object, keeps track of loss history, writes solutions to files, writes best/summary file
    sv = saver.saver(config=cnfg, name_config_file=filename, dim=dimensions)
    # iterate over number samples
    for i in range(cnfg['samples']):
        if i == 0:
            seed = cnfg['seed']
        else:
            random.seed()
            seed = random.randrange(1, 2 ** 32 - 1)

        random.seed(seed)
        np.random.seed(seed=seed)
        cnfg['seed'] = seed

        # initialize optimizer object, do preoptimization on complete graph, truncate graph according to bulk_thr
        optimizer = topological_opti(start_graph, sv, ent_dic=sys_dict, target_state=target_state, config=cnfg)
        if cnfg['topopt']:
            # topological optimization (deleting edges one by one)
            graph_res = optimizer.topologicalOptimization()
        else:
            # if topological optimization not wanted, return graph after optimization on complete graph
            graph_res = optimizer.graph
        # write solution to file
        sv.save_graph(optimizer)
        # compute state of result graph
        graph_res.getState()
        print(f'finished with graph with {len(graph_res.edges)} edges.')
        print(graph_res.state.state)
    return graph_res


def setup_for_fockbasis(cnfg):
    try:
        if cnfg["amplitudes"]:
            print('amplitudes = ', cnfg["amplitudes"])
        else:
            print('amplitudes left empty, assuming constant values of one')
    except KeyError:
        print('amplitudes not given, assuming constant values of one')
        cnfg["amplitudes"] = []

    try:
        if cnfg["imaginary"]:
            print('imaginary given: ', cnfg["imaginary"])
        else:
            print('real numbers used')
    except KeyError:
        print('imaginary not given, assuming real numbers.')
        cnfg["imaginary"] = False

    sys_dict = None

    # term_list = [term + cnfg['num_anc'] * '1' for term in cnfg["target_state"]]
    term_list = []
    for term in cnfg["target_state"]:
        ket = []
        for ii, tt in enumerate(term):
            ket += [(ii, 0)] * tt
        # ket = [ for ii, tt in enumerate(term)]
        for ii in range(cnfg['num_anc']):
            ket.append((len(term) + ii, 0))
        term_list.append(tuple(ket))
    # print(np.shape(term_list))
    num_out = len(cnfg["target_state"][0])
    cnfg["out_nodes"] = list(range(num_out))
    cnfg["in_nodes"] = []
    cnfg["single_emitters"] = []
    cnfg["verts"] = list(range(num_out + cnfg["num_anc"]))

    # not the corrected target_state but has been modified in the loss function
    # this can be changed afterwards
    target_state = State(term_list, amplitudes=cnfg['amplitudes'], imaginary=cnfg['imaginary'])

    # print(hf.readableState(target_state))
    num_mode_particle = len(cnfg["target_state"][0])
    dimensions = [1] * (num_mode_particle + cnfg['num_anc'])  # only one dimension at the moment

    edge_list = th.buildAllEdges(dimensions, imaginary=cnfg["imaginary"], loops=cnfg["loops"])
    edge_list = hf.prepEdgeList(edge_list, cnfg)
    print(f'start graph has {len(edge_list)} edges.')
    start_graph = Graph(edge_list, imaginary=cnfg['imaginary'])
    print(dimensions)
    return target_state, dimensions, sys_dict, start_graph


def setup_for_ent(cnfg):
    # concurrence optimization
    # define local dimensions
    dimensions = [int(ii) for ii in str(cnfg['dim'])]
    cnfg['dimensions'] = dimensions
    if len(dimensions) % 2 != 0:
        dimensions.append(1)
    target_state = None
    # compute sys_dict
    sys_dict = hf.get_sysdict(dimensions, bipar_for_opti=cnfg['K'], imaginary=cnfg['imaginary'])
    # build starting graph
    edge_list = th.buildAllEdges(dimensions, imaginary=cnfg['imaginary'])
    edge_list = hf.prepEdgeList(edge_list, cnfg)
    print(f'start graph has {len(edge_list)} edges.')
    start_graph = Graph(edge_list, imaginary=cnfg['imaginary'])
    return dimensions, sys_dict, start_graph


def setup_for_target(cnfg, state_cat=True):
    # default values
    try:
        cnfg["in_nodes"]
    except KeyError:
        cnfg["in_nodes"] = []

    try:
        cnfg["out_nodes"]
    except KeyError:
        cnfg["out_nodes"] = []

    if not cnfg["in_nodes"]:
        if not cnfg["out_nodes"]:
            print('no in/out nodes given. assuming that target terms correspond to out_nodes. state creation mode')
            cnfg["out_nodes"] = list(range(len(cnfg["target_state"][0])))
        else:
            print('no in_nodes given. state creation mode.')
    else:
        if not cnfg["out_nodes"]:
            print('no out_nodes given. assuming that target terms correspond to in_nodes. measurement mode')
        else:
            print('in_nodes and out_nodes given. target terms are read as a logic table for a quantum gate')

    if len(cnfg["out_nodes"]) + len(cnfg["in_nodes"]) != len(cnfg["target_state"][0]):
        print('TARGET DOES NOT MATCH IN_NODES AND cnfg.out_nodes')

    # num_anc gives the number of photons that go into detectors that are not cnfg.out_nodes (including those coming from single photon sources)
    try:
        print(f'number of ancillary photons = {cnfg["num_anc"]}')
    except KeyError:
        print('num_anc not given, assuming that number of ancillary photons = 0')
        cnfg["num_anc"] = 0

    try:
        if cnfg["single_emitters"]:
            print('single_emitters given. nodes corresponding to single photon sources: ', cnfg["single_emitters"])
        else:
            print('single_emitters not given. no single photon sources in setup.')
    except KeyError:
        print('no single photon emitters used')
        cnfg["single_emitters"] = []

    # add num_anc+len(single_emitters) vertices to graph (every ancillary detector and every single emitter needs a node)
    if cnfg["num_anc"] + len(cnfg["out_nodes"]) < len(cnfg["in_nodes"]) + len(cnfg["single_emitters"]):
        print('not enough ancillas given')
    additional_nodes = cnfg["num_anc"] + len(cnfg["single_emitters"])
    if not cnfg["out_nodes"]:
        additional_nodes += len(cnfg["in_nodes"])

    try:
        if cnfg["removed_connections"]:
            print('removed_connections given. additional constraints on the graph.')
            print('removed_connections = ', cnfg["removed_connections"])
        else:
            print('removed_connections not given. no additional constraints on the graph')
    except KeyError:
        print('removed_connections not given. assuming no additional constraints')
        cnfg["removed_connections"] = []

    try:
        if cnfg["unicolor"]:
            print("unicolor simplification used.")
    except KeyError:
        cnfg["unicolor"] = False

    try:
        if cnfg["amplitudes"]:
            print('amplitudes = ', cnfg["amplitudes"])
        else:
            print('amplitudes left empty, assuming constant values of one')
    except KeyError:
        print('amplitudes not given, assuming constant values of one')
        cnfg["amplitudes"] = []

    try:
        if cnfg["imaginary"]:
            print('imaginary given: ', cnfg["imaginary"])
        else:
            print('real numbers used')
    except KeyError:
        print('imaginary not given, assuming real numbers.')
        cnfg["imaginary"] = False

    try:
        if cnfg["heralding_out"]:
            print("heralding_out = True. out_nodes are not detected. ancillary detectors herald the outgoing state")
        else:
            print("heralding_out = False. out_nodes are detected. outgoing state is post-selected.")
    except KeyError:
        print('heralding_out not given, assuming post-selection')
        cnfg["heralding_out"] = False

    try:
        if cnfg["number_resolving"]:
            print('number resolving detectors used.')
        else:
            print('no number resolving detectors used')
    except KeyError:
        print('no information about photon-number resolving detectors given, assuming none are used')
        cnfg["number_resolving"] = False

    try:
        cnfg["brutal_covers"]
    except KeyError:
        cnfg["brutal_covers"] = False

    try:
        cnfg["bulk_thr"]
    except KeyError:
        cnfg["bulk_thr"] = 0

    try:
        cnfg["save_hist"]
    except KeyError:
        cnfg["save_hist"] = True

    try:
        cnfg["num_pre"]
    except KeyError:
        cnfg["num_pre"] = 1

    # define target
    target = [term + additional_nodes * '0' for term in cnfg["target_state"]]
    target_state = State(target, amplitudes=cnfg["amplitudes"], imaginary=cnfg["imaginary"])
    # print readable expression of the target state
    print(hf.readableState(target_state))

    # build starting graph
    # local dimensions necessary for each node to produce target
    cnfg["dimensions"] = th.stateDimensions(target_state.kets)
    # get complete starting graph according to local dimensions
    edge_list = th.buildAllEdges(cnfg["dimensions"], imaginary=cnfg["imaginary"])

    # introduce topological constraints
    # start with explicitly removed connections
    removed_connections = cnfg["removed_connections"]
    # add other restrictions imposed by specific kinds of nodes
    disjoint_nodes = cnfg["single_emitters"] + cnfg["in_nodes"]
    removed_connections += [sorted(con) for con in list(itertools.combinations(disjoint_nodes, 2))]
    edge_list = hf.removeConnections(edge_list, removed_connections)
    # apply unicolor simplification
    if cnfg['unicolor']:
        num_data_nodes = len(cnfg['target_state'][0])
        edge_list = hf.makeUnicolor(edge_list, num_data_nodes)
    print(f'start graph has {len(edge_list)} edges.')

    cnfg["verts"] = np.unique(list(itertools.chain(*th.edgeBleach(edge_list).keys())))
    cnfg["anc_detectors"] = [ii for ii in cnfg["verts"] if
                             ii not in cnfg["out_nodes"] + cnfg["single_emitters"] + cnfg["in_nodes"]]

    # turn edge list into graph
    graph = Graph(edge_list, imaginary=cnfg["imaginary"], state_cat=state_cat)

    return target_state, graph, cnfg


def build_starting_graph(cnfg, dimensions):
    # build starting graph
    edge_list = th.buildAllEdges(dimensions, imaginary=cnfg['imaginary'])
    edge_list = hf.prepEdgeList(edge_list, cnfg)
    print(f'start graph has {len(edge_list)} edges.')
    start_graph = Graph(edge_list, imaginary=cnfg['imaginary'])
    return start_graph


def get_dimensions_and_target_state(cnfg):
    # check if we are optimizing for entanglement
    # there is not a concrete target state and you have to define sys_dict
    if cnfg['loss_func'] == 'ent':
        # concurrence optimization
        # define local dimensions
        dimensions = [int(ii) for ii in str(cnfg['dim'])]
        if len(dimensions) % 2 != 0:
            dimensions.append(1)
        target_state = None
        # compute sys_dict
        sys_dict = hf.get_sysdict(dimensions, bipar_for_opti=cnfg['K'],
                                  imaginary=cnfg['imaginary'])
    else:
        # target state optimization
        sys_dict = None
        # add ancillas
        term_list = [term + cnfg['num_anc'] * '0' for term in cnfg['target_state']]
        # include amplitudes in target state if given
        if 'amplitudes' in cnfg:
            target_state = State(term_list, amplitudes=cnfg['amplitudes'], imaginary=cnfg['imaginary'])
        else:
            target_state = State(term_list, imaginary=cnfg['imaginary'])
        # print readable expression of the target state
        print(hf.readableState(target_state))
        target_kets = target_state.kets
        # define local dimensions
        dimensions = th.stateDimensions(target_kets)
    return dimensions, sys_dict, target_state


def read_config(is_example, filename):
    ''''
    read config json and output cnfg dict
    '''

    # option for running files from example folder
    if is_example:
        examples_dir = pkg_resources.resource_filename(pytheus.__name__, "graphs")
        for root, subdirs, files in os.walk(examples_dir):
            if root.split('/')[-1] == filename:
                for file in files:
                    if file.startswith('config'):
                        filename = root+'/'+file
                        break

    # check if filename ends in json, add extension if needed
    if not filename.endswith('.json'):
        filename += '.json'

    # error if file does not exist
    if not os.path.exists(filename) or os.path.isdir(filename):
        raise IOError(f'File does not exist: {filename}')
    # load json into dict
    with open(filename) as input_file:
        cnfg = json.load(input_file)
    # set some default value for some keys of dict
    if 'topopt' not in cnfg:
        cnfg['topopt'] = True
    if 'seed' not in cnfg:
        cnfg['seed'] = random.randrange(1, 2 ** 32 - 1)
    if not cnfg['topopt']:
        cnfg['bulk_thr'] = 0
    return cnfg, filename
